import base64
import ctypes
import datetime
import hashlib
import ipaddress
import os
import platform
import random
import re
import shutil
import socket
import ssl
import stat
import time
import traceback
import urllib.request
import uuid
import zipfile
from contextlib import closing
from itertools import cycle
from urllib.parse import urljoin
from xml.etree import ElementTree

import certifi
import cloudscraper
import ifaddr
import rarfile
import requests
import urllib3.exceptions
from cachecontrol import CacheControl
from cloudscraper.exceptions import CloudflareException
from tornado._locale_data import LOCALE_NAMES
from unidecode import unidecode
from urllib3 import disable_warnings

import sickchill
from sickchill import adba, logger, settings
from sickchill.helper import episode_num, MEDIA_EXTENSIONS, pretty_file_size, SUBTITLE_EXTENSIONS
from sickchill.helper.common import replace_extension
from sickchill.oldbeard.common import USER_AGENT
from sickchill.show.Show import Show

from . import db

# Add some missing languages
LOCALE_NAMES.update(
    {
        "ar_SA": {"name_en": "Arabic (Saudi Arabia)", "name": "(العربية (المملكة العربية السعودية"},
        "no_NO": {"name_en": "Norwegian", "name": "Norsk"},
    }
)


_context = None


def make_context(verify: bool):
    global _context
    if not _context or (_context and _context.check_hostname != verify):
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        context.options |= ssl.OP_NO_SSLv2
        context.verify_mode = ssl.CERT_REQUIRED if verify else ssl.CERT_NONE
        context.check_hostname = verify
        context.load_verify_locations(certifi.where(), None)
        _context = context
    return _context


def set_opener(verify: bool):
    disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    try:
        from urllib.request import HTTPSHandler

        https_handler = HTTPSHandler(context=make_context(verify), check_hostname=True)
        opener = urllib.request.build_opener(https_handler)
    except ImportError:
        opener = urllib.request.build_opener()

    opener.addheaders = [("User-agent", sickchill.oldbeard.common.USER_AGENT)]
    urllib.request.install_opener(opener)


set_opener(settings.SSL_VERIFY)

# Override original shutil function to increase its speed by increasing its buffer to 10MB (optimal)
copyfileobj_orig = shutil.copyfileobj


def _copyfileobj(fsrc, fdst, length=10485760):
    """ Run shutil.copyfileobj with a bigger buffer """
    return copyfileobj_orig(fsrc, fdst, length)


shutil.copyfileobj = _copyfileobj


def indentXML(elem, level=0):
    """
    Does our pretty printing, makes Matt very happy
    """
    i = "\n" + level * "  "
    if elem:
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indentXML(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def remove_non_release_groups(name):
    """
    Remove non release groups from name
    """
    if not name:
        return name

    # Do not remove all [....] suffixes, or it will break anime releases ## Need to verify this is true now
    # Check your database for funky release_names and add them here, to improve failed handling, archiving, and history.
    # select release_name from tv_episodes WHERE LENGTH(release_name);
    # [eSc], [SSG], [GWC] are valid release groups for non-anime
    removeWordsList = {
        r"\[rartv\]$": "searchre",
        r"\[rarbg\]$": "searchre",
        r"\.\[eztv\]$": "searchre",
        r"\[eztv\]$": "searchre",
        r"\[ettv\]$": "searchre",
        r"\[cttv\]$": "searchre",
        r"\.\[vtv\]$": "searchre",
        r"\[vtv\]$": "searchre",
        r"\[EtHD\]$": "searchre",
        r"\[GloDLS\]$": "searchre",
        r"\[silv4\]$": "searchre",
        r"\[Seedbox\]$": "searchre",
        r"\[PublicHD\]$": "searchre",
        r"\.\[PublicHD\]$": "searchre",
        r"\.\[NO.RAR\]$": "searchre",
        r"\[NO.RAR\]$": "searchre",
        r"-\=\{SPARROW\}\=-$": "searchre",
        r"\=\{SPARR$": "searchre",
        r"\.\[720P\]\[HEVC\]$": "searchre",
        r"\[AndroidTwoU\]$": "searchre",
        r"\[brassetv\]$": "searchre",
        r"\[Talamasca32\]$": "searchre",
        r"\(musicbolt\.com\)$": "searchre",
        r"\.\(NLsub\)$": "searchre",
        r"\(NLsub\)$": "searchre",
        r"\.\[BT\]$": "searchre",
        r" \[1044\]$": "searchre",
        r"\.RiPSaLoT$": "searchre",
        r"\.GiuseppeTnT$": "searchre",
        r"\.Renc$": "searchre",
        r"\.gz$": "searchre",
        r"\.English$": "searchre",
        r"\.German$": "searchre",
        r"\.\.Italian$": "searchre",
        r"\.Italian$": "searchre",
        r"(?<![57])\.1$": "searchre",
        r"-NZBGEEK$": "searchre",
        r"-Siklopentan$": "searchre",
        r"-Chamele0n$": "searchre",
        r"-Obfuscated$": "searchre",
        r"-Pre$": "searchre",
        r"-postbot$": "searchre",
        r"-BUYMORE$": "searchre",
        r"-\[SpastikusTV\]$": "searchre",
        r"-RP$": "searchre",
        r"-20-40$": "searchre",
        r"\.\[www\.usabit\.com\]$": "searchre",
        r"^\[www\.Cpasbien\.pe\] ": "searchre",
        r"^\[www\.Cpasbien\.com\] ": "searchre",
        r"^\[ www\.Cpasbien\.pw \] ": "searchre",
        r"^\.www\.Cpasbien\.pw": "searchre",
        r"^\[www\.newpct1\.com\]": "searchre",
        r"^\[ www\.Cpasbien\.com \] ": "searchre",
        r"- \{ www\.SceneTime\.com \}$": "searchre",
        r"^\{ www\.SceneTime\.com \} - ": "searchre",
        r"^\]\.\[www\.tensiontorrent.com\] - ": "searchre",
        r"^\]\.\[ www\.tensiontorrent.com \] - ": "searchre",
        r"- \[ www\.torrentday\.com \]$": "searchre",
        r"^\[ www\.TorrentDay\.com \] - ": "searchre",
        r"\[NO-RAR\] - \[ www\.torrentday\.com \]$": "searchre",
        r"^www\.Torrenting\.com\.-\.": "searchre",
        r"-Scrambled$": "searchre",
        r"^Torrent9\.PH ---> ": "searchre",
        r"-xpost$": "searchre",
    }

    _name = name
    for remove_string, remove_type in removeWordsList.items():
        if remove_type == "search":
            _name = _name.replace(remove_string, "")
        elif remove_type == "searchre":
            _name = re.sub(r"(?i){}".format(remove_string), "", _name)

    return _name.strip()


def is_media_file(filename):
    """
    Check if named file may contain media

    :param filename: Filename to check
    :return: True if this is a known media file, False if not
    """

    # ignore samples
    try:
        assert isinstance(filename, str), type(filename)
        is_rar = is_rar_file(filename)
        filename = os.path.basename(filename)

        if re.search(r"(^|[\W_])(?<!shomin.)(sample\d*)[\W_]", filename, re.I):
            return False

        # ignore RARBG release intro
        if re.search(r"^RARBG\.(\w+\.)?(mp4|avi|txt)$", filename, re.I):
            return False

        # ignore Kodi tvshow trailers
        if filename == "tvshow-trailer.mp4":
            return False

        # ignore MAC OS's retarded "resource fork" files
        if filename.startswith("._"):
            return False

        filname_parts = filename.rpartition(".")

        if re.search("extras?$", filname_parts[0], re.I):
            return False

        return filname_parts[-1].lower() in MEDIA_EXTENSIONS or (settings.UNPACK == settings.UNPACK_PROCESS_INTACT and is_rar)
    except (TypeError, AssertionError) as error:  # Not a string
        logger.debug(_("Invalid filename. Filename must be a string. {0}").format(error))
        return False


def is_rar_file(filename):
    """
    Check if file is a RAR file, or part of a RAR set

    :param filename: Filename to check
    :return: True if this is RAR/Part file, False if not
    """
    archive_regex = r"(?P<file>^(?P<base>(?:(?!\.part\d+\.rar$).)*)\.(?:(?:part0*1\.)?rar)$)"
    ret = re.search(archive_regex, filename) is not None
    try:
        if ret and os.path.exists(filename) and os.path.isfile(filename):
            ret = rarfile.is_rarfile(filename)
    except (IOError, OSError):
        pass

    return ret


def remove_file_failed(failed_file):
    """
    Remove file from filesystem

    :param failed_file: File to remove
    """

    # noinspection PyBroadException
    try:
        os.remove(failed_file)
    except Exception:
        pass


def makeDir(path):
    """
    Make a directory on the filesystem

    :param path: directory to make
    :return: True if success, False if failure
    """

    if not os.path.isdir(path):
        try:
            os.makedirs(path)
            # do the library update for synoindex
            sickchill.oldbeard.notifiers.synoindex_notifier.addFolder(path)
        except OSError:
            return False
    return True


def list_media_files(path):
    """
    Get a list of files possibly containing media in a path

    :param path: Path to check for files
    :return: list of files
    """

    if not path or not os.path.isdir(path):
        return []

    files = []
    for entry in os.listdir(path):
        full_entry = os.path.join(path, entry)

        # if it's a folder do it recursively
        if os.path.isdir(full_entry) and not entry.startswith(".") and not entry == "Extras":
            files += list_media_files(full_entry)

        elif is_media_file(entry):
            files.append(full_entry)

    return files


def copyFile(srcFile, destFile):
    """
    Copy a file from source to destination

    :param srcFile: Path of source file
    :param destFile: Path of destination file
    """

    try:
        shutil.copyfile(srcFile, destFile)
    except shutil.SameFileError:
        return
    except (shutil.SpecialFileError, shutil.Error) as error:
        logger.warning("{0}".format(error))
        raise error
    except Exception as error:
        logger.exception("{0}".format(error))
        raise error

    try:
        shutil.copymode(srcFile, destFile)
    except OSError:
        pass


def moveFile(srcFile, destFile):
    """
    Move a file from source to destination

    :param srcFile: Path of source file
    :param destFile: Path of destination file
    """
    try:
        shutil.move(srcFile, destFile)
    except OSError:
        copyFile(srcFile, destFile)
        os.unlink(srcFile)

    fixSetGroupID(destFile)


def hardlinkFile(srcFile, destFile):
    """
    Create a hard-link (inside filesystem link) between source and destination

    :param srcFile: Source file
    :param destFile: Destination file
    """

    try:
        os.link(srcFile, destFile)
    except Exception as error:
        logger.warning(_("Failed to create hardlink of {0} at {1}. Error: {2}. Copying instead").format(srcFile, destFile, error))
        copyFile(srcFile, destFile)

    fixSetGroupID(destFile)


def moveAndSymlinkFile(srcFile, destFile):
    """
    Move a file from source to destination, then create a symlink back from destination from source. If this fails, copy
    the file from source to destination

    :param srcFile: Source file
    :param destFile: Destination file
    """

    try:
        moveFile(srcFile, destFile)
        os.symlink(destFile, srcFile)
    except Exception as error:
        logger.warning(_("Failed to create symlink of {0} at {1}. Error: {2}. Copying instead").format(srcFile, destFile, error))
        copyFile(srcFile, destFile)


def make_dirs(path):
    """
    Creates any folders that are missing and assigns them the permissions of their
    parents
    """

    logger.debug(_("Checking if the path {0} already exists").format(path))

    if not os.path.isdir(path):
        # Windows, create all missing folders
        if platform.system() == "Windows":
            try:
                logger.debug(_("Folder {0} didn't exist, creating it").format(path))
                os.makedirs(path)
            except (OSError, IOError) as error:
                logger.exception(_("Failed creating {0} : {1}").format(path, error))
                return False

        # not Windows, create all missing folders and set permissions
        else:
            sofar = ""
            folder_list = path.split(os.path.sep)

            # look through each subfolder and make sure they all exist
            for cur_folder in folder_list:
                sofar += cur_folder + os.path.sep

                # if it exists then just keep walking down the line
                if os.path.isdir(sofar):
                    continue

                try:
                    logger.debug(_("Folder {0} didn't exist, creating it").format(sofar))
                    os.mkdir(sofar)
                    # use normpath to remove end separator, otherwise checks permissions against itself
                    chmodAsParent(os.path.normpath(sofar))
                    # do the library update for synoindex
                    sickchill.oldbeard.notifiers.synoindex_notifier.addFolder(sofar)
                except (OSError, IOError) as error:
                    logger.exception(_("Failed creating {0} : {1}").format(sofar, error))
                    return False

    return True


def rename_ep_file(cur_path, new_path, old_path_length=0):
    """
    Creates all folders needed to move a file to its new location, renames it, then cleans up any folders
    left that are now empty.

    :param  cur_path: The absolute path to the file you want to move/rename
    :param new_path: The absolute path to the destination for the file WITHOUT THE EXTENSION
    :param old_path_length: The length of media file path (old name) WITHOUT THE EXTENSION
    """

    # new_dest_dir, new_dest_name = os.path.split(new_path)

    if old_path_length == 0 or old_path_length > len(cur_path):
        # approach from the right
        cur_filename, cur_file_ext = os.path.splitext(cur_path)
    else:
        # approach from the left
        cur_file_ext = cur_path[old_path_length:]
        cur_filename = cur_path[:old_path_length]

    if cur_file_ext[1:] in SUBTITLE_EXTENSIONS:
        # Extract subtitle language from filename
        sublang = os.path.splitext(cur_filename)[1][1:]

        # Check if the language extracted from filename is a valid language
        if sublang in sickchill.oldbeard.subtitles.subtitle_code_filter():
            cur_file_ext = "." + sublang + cur_file_ext

    # put the extension on the incoming file
    new_path += cur_file_ext

    make_dirs(os.path.dirname(new_path))

    # move the file
    try:
        logger.info(_("Renaming file from {0} to {1}").format(cur_path, new_path))
        shutil.move(cur_path, new_path)
    except (OSError, IOError) as error:
        logger.exception(_("Failed renaming {0} to {1} : {2}").format(cur_path, new_path, error))
        return False

    # clean up any old folders that are empty
    delete_empty_folders(os.path.dirname(cur_path))

    return True


def delete_empty_folders(check_empty_dir, keep_dir=None):
    """
    Walks backwards up the path and deletes any empty folders found.

    :param check_empty_dir: The path to clean (absolute path to a folder)
    :param keep_dir: Clean until this path is reached
    """

    # treat check_empty_dir as empty when it only contains these items
    ignore_items = []

    logger.info(_("Trying to clean any empty folders under ") + check_empty_dir)

    # as long as the folder exists and doesn't contain any files, delete it
    while os.path.isdir(check_empty_dir) and check_empty_dir != keep_dir:
        check_files = os.listdir(check_empty_dir)

        if not check_files or (len(check_files) <= len(ignore_items) and all(check_file in ignore_items for check_file in check_files)):
            # directory is empty or contains only ignore_items
            try:
                logger.info(_("Deleting empty folder: ") + check_empty_dir)
                # need shutil.rmtree when ignore_items is really implemented
                os.rmdir(check_empty_dir)
                # do the library update for synoindex
                sickchill.oldbeard.notifiers.synoindex_notifier.deleteFolder(check_empty_dir)
            except OSError as error:
                logger.warning(_("Unable to delete {0}. Error: {1}").format(check_empty_dir, error))
                break
            check_empty_dir = os.path.dirname(check_empty_dir)
        else:
            break


def fileBitFilter(mode):
    """
    Strip special filesystem bits from file

    :param mode: mode to check and strip
    :return: required mode for media file
    """

    for bit in [stat.S_IXUSR, stat.S_IXGRP, stat.S_IXOTH, stat.S_ISUID, stat.S_ISGID]:
        if mode & bit:
            mode -= bit

    return mode


def chmodAsParent(childPath):
    """
    Retain permissions of parent for childs
    (Does not work for Windows hosts)

    :param childPath: Child Path to change permissions to sync from parent
    """

    if platform.system() == "Windows":
        return

    parentPath = os.path.dirname(childPath)

    if not parentPath:
        logger.debug(_("No parent path provided in {location} unable to get permissions from it").format(location=childPath))
        return

    childPath = os.path.join(parentPath, os.path.basename(childPath))

    parentPathStat = os.stat(parentPath)
    parentMode = stat.S_IMODE(parentPathStat[stat.ST_MODE])

    childPathStat = os.stat(childPath)
    childPath_mode = stat.S_IMODE(childPathStat[stat.ST_MODE])

    if os.path.isfile(childPath):
        childMode = fileBitFilter(parentMode)
    else:
        childMode = parentMode

    if childPath_mode == childMode:
        return

    childPath_owner = childPathStat.st_uid
    user_id = os.geteuid()  # @UndefinedVariable - only available on UNIX

    if user_id not in (childPath_owner, 0):
        logger.debug(_("Not running as root or owner of {location}, not trying to set permissions").format(location=childPath))
        return

    try:
        os.chmod(childPath, childMode)
    except OSError:
        logger.debug(_("Failed to set permission for {0} to {1:o}, parent directory has {2:o}").format(childPath, childMode, parentMode))


def fixSetGroupID(childPath):
    """
    Inherid SGID from parent
    (does not work on Windows hosts)

    :param childPath: Path to inherit SGID permissions from parent
    """

    if platform.system() == "Windows":
        return

    try:
        parentPath = os.path.dirname(childPath)
        parentStat = os.stat(parentPath)
        parentMode = stat.S_IMODE(parentStat[stat.ST_MODE])

        childPath = os.path.join(parentPath, os.path.basename(childPath))

        if parentMode & stat.S_ISGID:
            parentGID = parentStat[stat.ST_GID]
            childStat = os.stat(childPath)
            childGID = childStat[stat.ST_GID]

            if childGID == parentGID:
                return

            childPath_owner = childStat.st_uid
            user_id = os.geteuid()

            if user_id not in (childPath_owner, 0):
                logger.debug(_("Not running as root or owner of {}, not trying to set the set-group-ID").format(childPath))
                return

            try:
                os.chown(childPath, -1, parentGID)
                logger.debug(_("Respecting the set-group-ID bit on the parent directory for {0}").format(childPath))
            except (OSError, PermissionError):
                logger.debug("Failed to respect the set-group-ID bit on the parent directory for {0} (setting group ID {1})".format(childPath, parentGID))
    except Exception as error:
        logger.debug(f"Error setting set-group-id on the parent directory. {error}")


def is_anime_in_show_list():
    """
    Check if any shows in list contain anime

    :return: True if global showlist contains Anime, False if not
    """

    for show in settings.showList:
        if show.is_anime:
            return True
    return False


def update_anime_support():
    """Check if we need to support anime, and if we do, enable the feature"""

    settings.ANIMESUPPORT = is_anime_in_show_list()


def get_absolute_number_from_season_and_episode(show, season, episode):
    """
    Find the absolute number for a show episode

    :param show: Show object
    :param season: Season number
    :param episode: Episode number
    :return: The absolute number
    """

    absolute_number = None

    if season and episode:
        main_db_con = db.DBConnection()
        sql = "SELECT * FROM tv_episodes WHERE showid = ? and season = ? and episode = ?"
        sql_results = main_db_con.select(sql, [show.indexerid, season, episode])

        if len(sql_results) == 1:
            absolute_number = int(sql_results[0]["absolute_number"])
            logger.debug(
                _("Found absolute number {absolute} for show {show} {ep}").format(absolute=absolute_number, show=show.name, ep=episode_num(season, episode))
            )
        else:
            logger.debug(_("No entries for absolute number for show {show} {ep}").format(show=show.name, ep=episode_num(season, episode)))

    return absolute_number


def get_all_episodes_from_absolute_number(show, absolute_numbers, indexer_id=None):
    episodes = []
    season = None

    if len(absolute_numbers):
        if not show and indexer_id:
            show = Show.find(settings.showList, indexer_id)

        for absolute_number in absolute_numbers if show else []:
            ep = show.getEpisode(None, None, absolute_number=absolute_number)
            if ep:
                episodes.append(ep.episode)
                season = ep.season  # this will always take the last found season so eps that cross the season border are not handeled well

    return season, episodes


def sanitizeSceneName(name, anime=False):
    """
    Takes a show name and returns the "scenified" version of it.
    :param name: The name to sanitize
    :param anime: Some show have a ' in their name(Kuroko's Basketball) and is needed for search.
    :return: A string containing the scene version of the show name given.
    """

    # assert isinstance(name, unicode), name + ' is not unicode'

    if not name:
        return ""

    bad_chars = ",:()!?\u2019"
    if not anime:
        bad_chars += "'"

    # strip out any bad chars
    for x in bad_chars:
        name = name.replace(x, "")

    # tidy up stuff that doesn't belong in scene names
    name = name.replace("&", "and")
    name = re.sub(r"[- /]+", ".", name)
    name = re.sub(r"[.]+", ".", name)

    if name.endswith("."):
        name = name[:-1]

    return name


def create_https_certificates(ssl_cert, ssl_key):
    """
    Create self-signed HTTPS certificares and store in paths 'ssl_cert' and 'ssl_key'

    :param ssl_cert: Path of SSL certificate file to write
    :param ssl_key: Path of SSL keyfile to write
    :return: True on success, False on failure
    """

    # assert isinstance(ssl_key, unicode)
    # assert isinstance(ssl_cert, unicode)

    # noinspection PyBroadException
    try:
        from OpenSSL import crypto

        from sickchill.certgen import createCertificate, createCertRequest, createKeyPair, TYPE_RSA
    except Exception:
        logger.info(traceback.format_exc())
        logger.warning(_("pyopenssl module missing, please install for https access"))
        return False

    cakey = createKeyPair(TYPE_RSA, 4096)
    careq = createCertRequest(cakey, CN="Certificate Authority")
    params = {
        "req": careq,
        "issuerCert": careq,
        "issuerKey": cakey,
        "serial": int(time.time()),
        "notBefore": 0,
        "notAfter": 60 * 60 * 24 * 365 * 10,  # ten years
        "digest": "sha256",
    }
    # Create the CA Certificate
    params["issuerCert"] = createCertificate(**params)

    pkey = createKeyPair(TYPE_RSA, 4096)
    params["req"] = createCertRequest(pkey, CN="SickChill")
    cert = createCertificate(**params)

    # Save the key and certificate to disk
    # noinspection PyBroadException
    try:
        # Module has no member
        open(ssl_key, "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))
        open(ssl_cert, "wb").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
    except Exception as error:
        logger.info(traceback.format_exc())
        logger.warning(_("Error creating SSL key and certificate {error}").format(error))
        return False

    logger.info(_("Created https key: {ssl_key}").format(ssl_key=ssl_key))
    logger.info(_("Created https cert: {ssl_cert}").format(ssl_cert=ssl_cert))
    return True


def backupVersionedFile(old_file, version):
    """
    Back up an old version of a file

    :param old_file: Original file, to take a backup from
    :param version: Version of file to store in backup
    :return: True if success, False if failure
    """

    numTries = 0

    new_file = old_file + "." + "v" + str(version)

    while not os.path.isfile(new_file):
        if not os.path.isfile(old_file):
            logger.debug(_("Not creating backup, {0} doesn't exist").format(old_file))
            break

        try:
            logger.debug(_("Trying to back up {0} to {1}").format(old_file, new_file))
            shutil.copy(old_file, new_file)
            logger.debug(_("Backup done"))
            break
        except Exception as error:
            logger.warning(_("Error while trying to back up {0} to {1} : {2}").format(old_file, new_file, error))
            numTries += 1
            time.sleep(1)
            logger.debug(_("Trying again."))

        if numTries >= 10:
            logger.exception(_("Unable to back up {0} to {1} please do it manually.").format(old_file, new_file))
            return False

    return True


def restoreVersionedFile(backup_file, version):
    """
    Restore a file version to original state

    :param backup_file: File to restore
    :param version: Version of file to restore
    :return: True on success, False on failure
    """

    numTries = 0

    new_file, ext_ = os.path.splitext(backup_file)
    restore_file = new_file + "." + "v" + str(version)

    if not os.path.isfile(new_file):
        logger.debug(_("Not restoring, {0} doesn't exist").format(new_file))
        return False

    try:
        logger.debug(_("Trying to backup {0} to {1}.r{2} before restoring backup").format(new_file, new_file, version))

        shutil.move(new_file, new_file + "." + "r" + str(version))
    except Exception as error:
        logger.warning(_("Error while trying to backup DB file {0} before proceeding with restore: {1}").format(restore_file, error))
        return False

    while not os.path.isfile(new_file):
        if not os.path.isfile(restore_file):
            logger.debug(_("Not restoring, {0} doesn't exist").format(restore_file))
            break

        try:
            logger.debug(_("Trying to restore file {0} to {1}").format(restore_file, new_file))
            shutil.copy(restore_file, new_file)
            logger.debug(_("Restore done"))
            break
        except Exception as error:
            logger.warning(_("Error while trying to restore file {0}. Error: {1}").format(restore_file, error))
            numTries += 1
            time.sleep(1)
            logger.debug(_("Trying again. Attempt #: {0}").format(numTries))

        if numTries >= 10:
            logger.warning(_("Unable to restore file {0} to {1}").format(restore_file, new_file))
            return False

    return True


def get_lan_ip():
    """Returns IP of system"""

    # noinspection PyBroadException
    try:
        return [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][0]
    except Exception:
        return socket.gethostname()


def check_url(url):
    """
    Check if a URL exists without downloading the whole file.
    We only check the URL header.
    """
    try:
        requests.head(url, verify=False).raise_for_status()
    except Exception as error:
        # noinspection PyTypeChecker
        handle_requests_exception(error)
        return False
    return True


def anon_url(*url):
    """
    Return a URL string consisting of the Anonymous redirect URL and an arbitrary number of values appended.
    """
    return "" if None in url else "{0}{1}".format(settings.ANON_REDIRECT, "".join(str(s) for s in url))


"""
Encryption
==========
By Pedro Jose Pereira Vieito <pvieito@gmail.com> (@pvieito)

* If encryption_version==0 then return data without encryption
* The keys should be unique for each device

To add a new encryption_version:
  1) Code your new encryption_version
  2) Update the last encryption_version available in webserve.py
  3) Remember to maintain old encryption versions and key generators for retrocompatibility
"""

# Key Generators
unique_key1 = hex(uuid.getnode() ** 2)  # Used in encryption v1


# Encryption Functions
def encrypt(data: str, encryption_version=0, _decrypt=False):
    if not data:
        return ""

    if isinstance(data, str):
        data = data.encode()

    result = data
    secret_cycle = cycle((unique_key1, settings.ENCRYPTION_SECRET)[encryption_version > 1].encode())
    if encryption_version in (1, 2):
        if _decrypt:
            result = "".join(chr(x ^ y) for (x, y) in zip(base64.decodebytes(data), secret_cycle))
        else:
            result = base64.encodebytes("".join(chr(x ^ y) for (x, y) in zip(data, secret_cycle)).encode()).decode().strip()

    if isinstance(result, bytes):
        result = result.decode()

    return result


def decrypt(data, encryption_version=0):
    return encrypt(data, encryption_version, _decrypt=True)


def full_sanitizeSceneName(name):
    return re.sub("[. -]", " ", sanitizeSceneName(name)).lower().strip()


def _check_against_names(nameInQuestion, show, season=-1):
    showNames = []
    if season in [-1, 1]:
        showNames = [show.name]

    showNames.extend(sickchill.oldbeard.scene_exceptions.get_scene_exceptions(show.indexerid, season=season))

    for showName in showNames:
        nameFromList = full_sanitizeSceneName(showName)
        if nameFromList == nameInQuestion:
            return True

    return False


def get_show(name, tryIndexers=False):
    if not settings.showList:
        return

    showObj = None
    fromCache = False

    if not name:
        return showObj

    try:
        # check cache for show
        cache = sickchill.oldbeard.name_cache.get_id_from_name(name)
        if cache:
            fromCache = True
            showObj = Show.find(settings.showList, int(cache))
        else:
            check_names = [full_sanitizeSceneName(name), name]
            show_matches = [
                show
                for show in settings.showList
                if (show.show_name and full_sanitizeSceneName(show.show_name) in check_names)
                or (show.custom_name and full_sanitizeSceneName(show.custom_name) in check_names)
            ]

            if len(show_matches) == 1:
                showObj = show_matches[0]

        # try indexers
        if not showObj and tryIndexers:
            showObj = Show.find(settings.showList, sickchill.indexer.search_indexers_for_series_id(name=full_sanitizeSceneName(name))[1].id)

        # try scene exceptions
        if not showObj:
            scene_exceptions = sickchill.oldbeard.scene_exceptions.get_scene_exception_by_name_multiple(name)
            for scene_exception in scene_exceptions:
                if scene_exception[0]:
                    showObj = Show.find(settings.showList, scene_exception[0])
                    if showObj:
                        break

        # add show to cache
        if showObj and not fromCache:
            sickchill.oldbeard.name_cache.add_name(name, showObj.indexerid)
    except Exception as error:
        logger.debug(_("Error when attempting to find show: {0} in SickChill. Error: {1} ").format(name, error))
        logger.debug(traceback.format_exc())

    return showObj


def is_hidden_folder(folder):
    """
    Returns True if folder is hidden.
    On Linux based systems hidden folders start with . (dot)
    :param folder: Full path of folder to check
    """

    def is_hidden(filepath):
        name = os.path.basename(os.path.abspath(filepath))
        return name == "@eaDir" or name.startswith(".") or has_hidden_attribute(filepath)

    def has_hidden_attribute(filepath):
        try:
            attrs = ctypes.windll.kernel32.GetFileAttributesW(ctypes.c_wchar_p(str(filepath)))
            assert attrs != -1
            result = bool(attrs & 2)
        except (AttributeError, AssertionError, OSError, IOError):
            result = False
        return result

    if os.path.isdir(folder) and is_hidden(folder):
        return True

    return False


def real_path(path):
    """
    Returns: the canonicalized absolute pathname. The resulting path will have no symbolic link, '/./' or '/../' components.
    """
    return os.path.normpath(os.path.normcase(os.path.realpath(path)))


def is_subdirectory(subdir_path, topdir_path):
    """
    Returns true if a subdir_path is a subdirectory of topdir_path
    else otherwise.

    :param subdir_path: The full path to the sub-directory
    :param topdir_path: The full path to the top directory to check subdir_path against
    """
    topdir_path = real_path(topdir_path)
    subdir_path = real_path(subdir_path)

    # checks if the common prefix of both is equal to directory
    # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    return os.path.commonprefix([subdir_path, topdir_path]) == topdir_path


def set_up_anidb_connection():
    """Connect to anidb"""

    if not settings.USE_ANIDB:
        logger.debug(_("Usage of anidb disabled. Skiping"))
        return False

    if not settings.ANIDB_USERNAME and not settings.ANIDB_PASSWORD:
        logger.debug(_("anidb username and/or password are not set. Aborting anidb lookup."))
        return False

    if not settings.ADBA_CONNECTION:

        def anidb_logger(msg):
            return logger.debug(_("anidb: {0} ").format(msg))

        try:
            settings.ADBA_CONNECTION = adba.Connection(keep_alive=True, log=anidb_logger)
        except Exception as error:
            logger.warning(_("anidb exception msg: {0} ").format(error))
            return False

    try:
        if not settings.ADBA_CONNECTION.authed():
            settings.ADBA_CONNECTION.auth(settings.ANIDB_USERNAME, settings.ANIDB_PASSWORD)
        else:
            return True
    except Exception as error:
        logger.warning(_("anidb exception msg: {0} ").format(error))
        return False

    return settings.ADBA_CONNECTION.authed()


def makeZip(fileList, archive):
    """
    Create a ZIP of files

    :param fileList: A list of file names - full path each name
    :param archive: File name for the archive with a full path
    """

    try:
        a = zipfile.ZipFile(archive, "w", zipfile.ZIP_DEFLATED, allowZip64=True)
        for f in fileList:
            a.write(f)
        a.close()
        return True
    except Exception as error:
        logger.exception(_("Zip creation error: {0} ").format(error))
        return False


def extractZip(archive, targetDir):
    """
    Unzip a file to a directory

    :param archive: The file name of the archive to extract with a full path
    :param targetDir: The full path to the extraction target directory
    """

    try:
        if not os.path.exists(targetDir):
            os.mkdir(targetDir)

        zip_file = zipfile.ZipFile(archive, "r", allowZip64=True)
        for member in zip_file.namelist():
            filename = os.path.basename(member)
            # skip directories
            if not filename:
                continue

            # copy file (taken from zipfile's extract)
            source = zip_file.open(member)
            target = open(os.path.join(targetDir, filename), "wb")
            shutil.copyfileobj(source, target)
            source.close()
            target.close()
        zip_file.close()
        return True
    except Exception as error:
        logger.exception(_("Zip extraction error: {0} ").format(error))
        return False


def backup_config_zip(fileList, archive, arcname=None):
    """
    Store the config file as a ZIP

    :param fileList: List of files to store
    :param archive: ZIP file name
    :param arcname: Archive path
    :return: True on success, False on failure
    """

    try:
        a = zipfile.ZipFile(archive, "w", zipfile.ZIP_DEFLATED)
        for f in fileList:
            a.write(f, os.path.relpath(f, arcname))
        a.close()
        return True
    except Exception as error:
        logger.warning(_("Zip creation error: {0} ").format(error))
        return False


def restore_config_zip(archive, targetDir):
    """
    Restores a Config ZIP file back in place

    :param archive: ZIP filename
    :param targetDir: Directory to restore to
    :return: True on success, False on failure
    """

    try:
        if not os.path.exists(targetDir):
            os.mkdir(targetDir)
        else:

            def path_leaf(path):
                head, tail = os.path.split(path)
                return tail or os.path.basename(head)

            bakFilename = "{0}-{1}".format(path_leaf(targetDir), datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
            shutil.move(targetDir, os.path.join(os.path.dirname(targetDir), bakFilename))

        zip_file = zipfile.ZipFile(archive, "r", allowZip64=True)
        for member in zip_file.namelist():
            zip_file.extract(member, targetDir)
        zip_file.close()
        return True
    except Exception as error:
        logger.exception(_("Zip extraction error: {0}").format(error))
        shutil.rmtree(targetDir)
        return False


def touchFile(fname, atime=None):
    """
    Touch a file (change modification date)

    :param fname: Filename to touch
    :param atime: Specific access time (defaults to None)
    :return: True on success, False on failure
    """

    if atime and fname and os.path.isfile(fname):
        os.utime(fname, (atime, atime))
        return True

    return False


def make_indexer_session():
    session = make_session()
    session.verify = (False, certifi.where())[settings.SSL_VERIFY]
    if settings.PROXY_SETTING and settings.PROXY_INDEXERS:
        logger.debug(_("Using global proxy for indexers: {}").format(settings.PROXY_SETTING))
        session.proxies = {"http": settings.PROXY_SETTING, "https": settings.PROXY_SETTING}
    return session


def make_session():
    session = requests.Session()
    session.headers.update({"User-Agent": USER_AGENT, "Accept-Encoding": "gzip,deflate"})
    session = cloudscraper.create_scraper(sess=session, delay=6)
    return CacheControl(sess=session, cache_etags=True)


def request_defaults(kwargs):
    verify = certifi.where() if all([settings.SSL_VERIFY, kwargs.pop("verify", True)]) else False

    # request session proxies
    if kwargs.pop("allow_proxy", True) and settings.PROXY_SETTING:
        logger.debug(_("Using global proxy: {}").format(settings.PROXY_SETTING))
        proxies = {"http": settings.PROXY_SETTING, "https": settings.PROXY_SETTING}
    else:
        proxies = None

    return kwargs.pop("hooks", None), kwargs.pop("cookies", None), verify, proxies


def getURL(
    url,
    post_data=None,
    params=None,
    headers=None,  # pylint:disable=too-many-arguments, too-many-return-statements, too-many-branches, too-many-locals
    timeout=30,
    session=None,
    **kwargs,
):
    """
    Returns data retrieved from the url provider.
    """
    try:
        response_type = kwargs.pop("returns", "text")
        stream = kwargs.pop("stream", False)
        allow_redirects = kwargs.pop("allow_redirects", True)

        hooks, cookies, verify, proxies = request_defaults(kwargs)

        resp = session.request(
            "POST" if post_data else "GET",
            url,
            data=post_data or {},
            params=params or {},
            timeout=timeout,
            allow_redirects=allow_redirects,
            hooks=hooks,
            stream=stream,
            headers=headers,
            cookies=cookies,
            proxies=proxies,
            verify=verify,
        )
        resp.raise_for_status()
    except Exception as error:
        # noinspection PyTypeChecker
        handle_requests_exception(error)
        return None

    try:
        return resp if response_type == "response" or response_type is None else resp.json() if response_type == "json" else getattr(resp, response_type, resp)
    except ValueError:
        logger.debug(_("Requested a json response but response was not json, check the url: {0}").format(url))
        return None


def download_file(url, filename, session=None, headers=None, **kwargs):  # pylint:disable=too-many-return-statements
    """
    Downloads a file specified

    :param url: Source URL
    :param filename: Target file on filesystem
    :param session: request session to use
    :param headers: override existing headers in request session
    :return: True on success, False on failure
    """

    return_filename = kwargs.get("return_filename", False)

    try:
        hooks, cookies, verify, proxies = request_defaults(kwargs)

        with closing(
            session.get(url, allow_redirects=True, stream=True, verify=verify, headers=headers, cookies=cookies, hooks=hooks, proxies=proxies)
        ) as resp:

            resp.raise_for_status()

            # Workaround for jackett.
            if filename.endswith("nzb") and resp.headers.get("content-type") == "application/x-bittorrent":
                filename = replace_extension(filename, "torrent")

            try:
                with open(filename, "wb") as fp:
                    for chunk in resp.iter_content(chunk_size=1024):
                        if chunk:
                            fp.write(chunk)
                            fp.flush()

                chmodAsParent(filename)
            except Exception as error:
                logger.warning(_("Problem downloading file, setting permissions or writing file to {0} - ERROR: {1}").format(filename, error))

    except Exception as error:
        # noinspection PyTypeChecker
        handle_requests_exception(error)
        return False if not return_filename else ""

    return True if not return_filename else filename


def handle_requests_exception(requests_exception):
    def get_level(exception):
        return (logger.ERROR, logger.WARNING)[exception and "s,t,o,p,b,r,e,a,k,i,n,g,f" in str(exception)]

    default = _("Request failed: {0} ({1})")
    try:
        raise requests_exception
    except requests.exceptions.SSLError as error:
        if ssl.OPENSSL_VERSION_INFO < (1, 0, 1, 5):
            logger.info(_("SSL Error requesting url: '{0}' You have {1}, try upgrading OpenSSL to 1.0.1e+").format(error.request.url, ssl.OPENSSL_VERSION))
        if settings.SSL_VERIFY:
            logger.info(_("SSL Error requesting url: '{0}' Try disabling Cert Verification on the advanced tab of /config/general").format(error.request.url))
        logger.debug(default.format(error, type(error.__class__.__name__)))
        logger.debug(traceback.format_exc())
    except requests.exceptions.ContentDecodingError as error:
        logger.info(default.format(error, type(error.__class__.__name__)))
        logger.debug(traceback.format_exc())
    except urllib3.exceptions.ProxySchemeUnknown as error:
        logger.info(default.format("You must prefix your proxy setting with a scheme (http/https/etc)", error))
    except CloudflareException as error:
        logger.info(default.format(error, type(error.__class__.__name__)))
    except requests.exceptions.RequestException as error:
        if not (
            hasattr(error, "response")
            and error.response
            and hasattr(error.response, "status_code")
            and error.response.status_code == 404
            and hasattr(error.response, "headers")
            and error.response.headers.get("X-Content-Type-Options") == "nosniff"
        ):
            logger.info(default.format(error, type(error.__class__.__name__)))
    except (TypeError, ValueError) as error:
        level = get_level(error)
        logger.log(level, default.format(error, type(error.__class__.__name__)))
        if requests_exception.request:
            logger.info(_("url is {0}").format(repr(requests_exception.request.url)))
            logger.info("headers are {0}".format(repr(requests_exception.request.headers)))
            logger.info("params are {0}".format(repr(requests_exception.request.params)))
            logger.info("post_data is {0}".format(repr(requests_exception.request.data)))
        if level == logger.WARNING:
            logger.debug(traceback.format_exc())
    except Exception as error:
        logger.log(get_level(error), default.format(error, type(error.__class__.__name__)))
        logger.debug(traceback.format_exc())


def get_size(start_path="."):
    """
    Find the total dir and filesize of a path

    :param start_path: Path to recursively count size
    :return: total filesize
    """

    if not os.path.isdir(start_path):
        return -1

    total_size = 0
    for dirpath, dirnames_, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if os.path.islink(fp) and not os.path.isfile(fp):
                log = logger.debug if settings.IGNORE_BROKEN_SYMLINKS else logger.warning
                log(_("Unable to get size for file {0} because the link to the file is not valid").format(fp))
                continue
            try:
                total_size += os.path.getsize(fp)
            except OSError as error:
                logger.exception(_("Unable to get size for file {0} Error: {1}").format(fp, error))
                logger.debug(traceback.format_exc())
    return total_size


def generateApiKey():
    """ Return a new randomized API_KEY"""
    logger.info(_("Generating New API key"))
    secure_hash = hashlib.sha512(str(time.time()).encode())
    secure_hash.update(str(random.SystemRandom().getrandbits(4096)).encode())
    return secure_hash.hexdigest()[:32]


def remove_article(text=""):
    """Remove the english articles from a text string"""

    return re.sub(r"(?i)^(?:(?:A(?!\s+to)n?)|The)\s(\w)", r"\1", text)


def generateCookieSecret():
    """Generate a new cookie secret"""

    return base64.b64encode(uuid.uuid4().bytes + uuid.uuid4().bytes).decode()


def disk_usage(path):
    if platform.system() == "Windows":
        free = ctypes.c_ulonglong(0)
        if ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(str(path)), None, None, ctypes.pointer(free)) == 0:
            raise ctypes.WinError()
        return free.value

    elif hasattr(os, "statvfs"):  # POSIX
        if platform.system() == "Darwin":
            try:
                import subprocess

                call = subprocess.Popen(["df", "-k", path], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
                output = call.communicate()[0]
                return int(output.split("\n")[1].split()[3]) * 1024
            except Exception:
                pass

        st = os.statvfs(path)
        return st.f_bavail * st.f_frsize

    else:
        raise Exception("Unable to determine free space on your OS")


def verify_freespace(src, dest, oldfile=None, method="copy"):
    """
    Checks if the target system has enough free space to copy or move a file.

    :param src: Source filename
    :param dest: Destination path (show dir in current usage)
    :param oldfile: File to be replaced (defaults to None)
    :param method: The post processing method
    :return: True if there is enough space for the file, False if there isn't. Also returns True if the OS doesn't support this option
    """

    if not isinstance(oldfile, list):
        oldfile = [oldfile] if oldfile else []

    logger.debug(_("Trying to determine free space on the destination drive"))

    if not os.path.isfile(src):
        logger.warning(_("A path to a file is required for the source. {0} is not a file.").format(src))
        return True

    if not (os.path.isdir(dest) or (settings.CREATE_MISSING_SHOW_DIRS and os.path.isdir(os.path.dirname(dest)))):
        logger.warning(
            _("A path is required for the destination. Check that the root dir and show locations are correct for {0} (I got '{1}')").format(
                oldfile[0].name, dest
            )
        )
        return False

    dest = (os.path.dirname(dest), dest)[os.path.isdir(dest)]

    # shortcut: if we are moving the file and the destination == src dir,
    # then by definition there is enough space
    if method == "move" and os.stat(src).st_dev == os.stat(dest).st_dev:  # pylint:
        # disable=no-member
        logger.info(_("Process method is 'move' and src and destination are on the same device, skipping free space check"))
        return True

    try:
        disk_free = disk_usage(dest)
    except Exception as error:
        logger.warning(_("Unable to determine free space, so I will assume there is enough."))
        logger.debug(_("Error: {error}").format(error=error))
        logger.debug(traceback.format_exc())
        return True

    # Lets also do this for symlink and hardlink
    if "link" in method and disk_free > 1024 ** 2:
        return True

    needed_space = os.path.getsize(src)

    if oldfile:
        for f in oldfile:
            if os.path.isfile(f.location):
                disk_free += os.path.getsize(f.location)

    if disk_free > needed_space:
        return True
    else:
        logger.warning(
            _("Not enough free space: Needed: {0} bytes ( {1} ), found: {2} bytes ( {3} )").format(
                needed_space, pretty_file_size(needed_space), disk_free, pretty_file_size(disk_free)
            )
        )
        return False


def disk_usage_hr(diskPath=None):
    """
    returns the free space in human readable bytes for a given path or False if no path given
    :param diskPath: the filesystem path being checked
    """
    if diskPath and os.path.exists(diskPath):
        try:
            free = disk_usage(diskPath)
        except Exception as error:
            logger.warning(_("Unable to determine free space"))
            logger.debug(_("Error: {error}").format(error=error))
            logger.debug(traceback.format_exc())
        else:
            return pretty_file_size(free)

    return False


# https://gist.github.com/thatalextaylor/7408395
def pretty_time_delta(seconds):
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    time_delta = sign_string

    if days > 0:
        time_delta += "{0}d".format(days)
    if hours > 0:
        time_delta += "{0}h".format(hours)
    if minutes > 0:
        time_delta += "{0}m".format(minutes)
    if seconds > 0:
        time_delta += "{0}s".format(seconds)

    return time_delta


def is_file_locked(checkfile, write_check=False):
    """
    Checks to see if a file is locked. Performs three checks
        1. Checks if the file even exists
        2. Attempts to open the file for reading. This will determine if the file has a write lock.
            Write locks occur when the file is being edited or copied to, e.g. a file copy destination
        3. If the readLockCheck parameter is True, attempts to rename the file. If this fails the
            file is open by some other process for reading. The file can be read, but not written to
            or deleted.
    :param checkfile: the file being checked
    :param write_check: when true will check if the file is locked for writing (prevents move operations)
    """

    checkfile = os.path.abspath(checkfile)

    if not os.path.exists(checkfile):
        return True
    try:
        f = open(checkfile, "rb")
        f.close()
    except IOError:
        return True

    if write_check:
        lockFile = checkfile + ".lckchk"
        if os.path.exists(lockFile):
            os.remove(lockFile)
        try:
            os.rename(checkfile, lockFile)
            time.sleep(1)
            os.rename(lockFile, checkfile)
        except (OSError, IOError):
            return True

    return False


def tvdbid_from_remote_id(indexer_id, indexer):  # pylint:disable=too-many-return-statements

    session = make_session()
    tvdb_id = ""
    if indexer == "IMDB":
        url = "http://www.thetvdb.com/api/GetSeriesByRemoteID.php?imdbid={0}".format(indexer_id)
        data = getURL(url, session=session, returns="content")
        if data is None:
            return tvdb_id
        try:
            tree = ElementTree.fromstring(data)
            for show in tree.iter("Series"):
                tvdb_id = show.findtext("seriesid")

        except SyntaxError:
            pass

        return tvdb_id
    elif indexer == "ZAP2IT":
        url = "http://www.thetvdb.com/api/GetSeriesByRemoteID.php?zap2it={0}".format(indexer_id)
        data = getURL(url, session=session, returns="content")
        if data is None:
            return tvdb_id
        try:
            tree = ElementTree.fromstring(data)
            for show in tree.iter("Series"):
                tvdb_id = show.findtext("seriesid")

        except SyntaxError:
            pass

        return tvdb_id
    elif indexer == "TVMAZE":
        url = "http://api.tvmaze.com/shows/{0}".format(indexer_id)
        data = getURL(url, session=session, returns="json")
        if data is None:
            return tvdb_id
        tvdb_id = data["externals"]["thetvdb"]
        return tvdb_id
    else:
        return tvdb_id


def is_ip_local(ip):
    request_ip = ipaddress.ip_address(ip)
    if request_ip.is_private:
        return True

    for adapter in ifaddr.get_adapters():
        for aip in adapter.ips:
            if isinstance(aip.ip, tuple):
                network = ipaddress.IPv6Network("%s/%s" % (aip.ip[0], aip.network_prefix), strict=False)
            else:
                network = ipaddress.IPv4Network("%s/%s" % (aip.ip, aip.network_prefix), strict=False)

            if request_ip in network:
                return True
    return False


MESSAGE_COUNTER = 0


def add_site_message(message, tag=None, level="danger"):
    with settings.MESSAGES_LOCK:
        to_add = dict(level=level, tag=tag, message=message)

        if tag:  # prevent duplicate messages of the same type
            # http://www.goodmami.org/2013/01/30/Getting-only-the-first-match-in-a-list-comprehension.html
            existing = next((x for x, msg in settings.SITE_MESSAGES.items() if msg.get("tag") == tag), None)
            if existing:
                settings.SITE_MESSAGES[existing] = to_add
                return

        global MESSAGE_COUNTER
        MESSAGE_COUNTER += 1
        settings.SITE_MESSAGES[MESSAGE_COUNTER] = to_add


def remove_site_message(key=None, tag=None):
    with settings.MESSAGES_LOCK:
        if key is not None and int(key) in settings.SITE_MESSAGES:
            del settings.SITE_MESSAGES[int(key)]
        elif tag is not None:
            found = [idx for idx, msg in settings.SITE_MESSAGES.items() if msg.get("tag") == tag]
            for key in found:
                del settings.SITE_MESSAGES[key]


def sortable_name(name):
    if not settings.SORT_ARTICLE:
        name = re.sub(r"(?:The|A|An)\s", "", name, flags=re.I)
    return unidecode(name.lower())


def manage_torrents_url(reset=False):
    if not reset:
        return settings.CLIENT_WEB_URLS.get("torrent", "")

    if (
        not settings.USE_TORRENTS
        or not settings.TORRENT_HOST.lower().startswith("http")
        or settings.TORRENT_METHOD == "blackhole"
        or settings.ENABLE_HTTPS
        and not settings.TORRENT_HOST.lower().startswith("https")
    ):
        settings.CLIENT_WEB_URLS["torrent"] = ""
        return settings.CLIENT_WEB_URLS.get("torrent")

    torrent_ui_url = re.sub("localhost|127.0.0.1", settings.LOCALHOST_IP or get_lan_ip(), settings.TORRENT_HOST or "", re.I)

    def test_exists(url):
        try:
            h = requests.head(url)
            return h.status_code != 404
        except requests.exceptions.RequestException:
            return False

    if settings.TORRENT_METHOD == "utorrent":
        torrent_ui_url = "/".join(s.strip("/") for s in (torrent_ui_url, "gui/"))
    elif settings.TORRENT_METHOD == "download_station":
        if test_exists(urljoin(torrent_ui_url, "download/")):
            torrent_ui_url = urljoin(torrent_ui_url, "download/")

    settings.CLIENT_WEB_URLS["torrent"] = ("", torrent_ui_url)[test_exists(torrent_ui_url)]

    return settings.CLIENT_WEB_URLS.get("torrent")


def is_docker():
    path = "/proc/self/cgroup"
    return os.path.exists("/.dockerenv") or os.path.isfile(path) and any("docker" in line for line in open(path))



# -*- coding: utf-8 -*-
import os
import sys
import codecs

try:
    from collections import UserList
except ImportError:
    from UserList import UserList

from itertools import chain
from copy import copy

from pysrt.srtexc import Error
from pysrt.srtitem import SubRipItem
from pysrt.compat import str

BOMS = ((codecs.BOM_UTF32_LE, 'utf_32_le'),
        (codecs.BOM_UTF32_BE, 'utf_32_be'),
        (codecs.BOM_UTF16_LE, 'utf_16_le'),
        (codecs.BOM_UTF16_BE, 'utf_16_be'),
        (codecs.BOM_UTF8, 'utf_8'))
CODECS_BOMS = dict((codec, str(bom, codec)) for bom, codec in BOMS)
BIGGER_BOM = max(len(bom) for bom, encoding in BOMS)


class SubRipFile(UserList, object):
    """
    SubRip file descriptor.

    Provide a pure Python mapping on all metadata.

    SubRipFile(items, eol, path, encoding)

    items -> list of SubRipItem. Default to [].
    eol -> str: end of line character. Default to linesep used in opened file
        if any else to os.linesep.
    path -> str: path where file will be saved. To open an existant file see
        SubRipFile.open.
    encoding -> str: encoding used at file save. Default to utf-8.
    """
    ERROR_PASS = 0
    ERROR_LOG = 1
    ERROR_RAISE = 2

    DEFAULT_ENCODING = 'utf_8'

    def __init__(self, items=None, eol=None, path=None, encoding='utf-8'):
        UserList.__init__(self, items or [])
        self._eol = eol
        self.path = path
        self.encoding = encoding

    def _get_eol(self):
        return self._eol or os.linesep

    def _set_eol(self, eol):
        self._eol = self._eol or eol

    eol = property(_get_eol, _set_eol)

    def slice(self, starts_before=None, starts_after=None, ends_before=None,
              ends_after=None):
        """
        slice([starts_before][, starts_after][, ends_before][, ends_after]) \
-> SubRipFile clone

        All arguments are optional, and should be coercible to SubRipTime
        object.

        It reduce the set of subtitles to those that match match given time
        constraints.

        The returned set is a clone, but still contains references to original
        subtitles. So if you shift this returned set, subs contained in the
        original SubRipFile instance will be altered too.

        Example:
            >>> subs.slice(ends_after={'seconds': 20}).shift(seconds=2)
        """
        clone = copy(self)

        if starts_before:
            clone.data = (i for i in clone.data if i.start < starts_before)
        if starts_after:
            clone.data = (i for i in clone.data if i.start > starts_after)
        if ends_before:
            clone.data = (i for i in clone.data if i.end < ends_before)
        if ends_after:
            clone.data = (i for i in clone.data if i.end > ends_after)

        clone.data = list(clone.data)
        return clone

    def at(self, timestamp=None, **kwargs):
        """
        at(timestamp) -> SubRipFile clone

        timestamp argument should be coercible to SubRipFile object.

        A specialization of slice. Return all subtiles visible at the
        timestamp mark.

        Example:
            >>> subs.at((0, 0, 20, 0)).shift(seconds=2)
            >>> subs.at(seconds=20).shift(seconds=2)
        """
        time = timestamp or kwargs
        return self.slice(starts_before=time, ends_after=time)

    def shift(self, *args, **kwargs):
        """shift(hours, minutes, seconds, milliseconds, ratio)

        Shift `start` and `end` attributes of each items of file either by
        applying a ratio or by adding an offset.

        `ratio` should be either an int or a float.
        Example to convert subtitles from 23.9 fps to 25 fps:
        >>> subs.shift(ratio=25/23.9)

        All "time" arguments are optional and have a default value of 0.
        Example to delay all subs from 2 seconds and half
        >>> subs.shift(seconds=2, milliseconds=500)
        """
        for item in self:
            item.shift(*args, **kwargs)

    def clean_indexes(self):
        """
        clean_indexes()

        Sort subs and reset their index attribute. Should be called after
        destructive operations like split or such.
        """
        self.sort()
        for index, item in enumerate(self):
            item.index = index + 1

    @property
    def text(self):
        return '\n'.join(i.text for i in self)

    @classmethod
    def open(cls, path='', encoding=None, error_handling=ERROR_PASS):
        """
        open([path, [encoding]])

        If you do not provide any encoding, it can be detected if the file
        contain a bit order mark, unless it is set to utf-8 as default.
        """
        source_file, encoding = cls._open_unicode_file(path, claimed_encoding=encoding)
        new_file = cls(path=path, encoding=encoding)
        new_file.read(source_file, error_handling=error_handling)
        source_file.close()
        return new_file

    @classmethod
    def from_string(cls, source, **kwargs):
        """
        from_string(source, **kwargs) -> SubRipFile

        `source` -> a unicode instance or at least a str instance encoded with
        `sys.getdefaultencoding()`
        """
        error_handling = kwargs.pop('error_handling', None)
        new_file = cls(**kwargs)
        new_file.read(source.splitlines(True), error_handling=error_handling)
        return new_file

    def read(self, source_file, error_handling=ERROR_PASS):
        """
        read(source_file, [error_handling])

        This method parse subtitles contained in `source_file` and append them
        to the current instance.

        `source_file` -> Any iterable that yield unicode strings, like a file
            opened with `codecs.open()` or an array of unicode.
        """
        self.eol = self._guess_eol(source_file)
        self.extend(self.stream(source_file, error_handling=error_handling))
        return self

    @classmethod
    def stream(cls, source_file, error_handling=ERROR_PASS):
        """
        stream(source_file, [error_handling])

        This method yield SubRipItem instances a soon as they have been parsed
        without storing them. It is a kind of SAX parser for .srt files.

        `source_file` -> Any iterable that yield unicode strings, like a file
            opened with `codecs.open()` or an array of unicode.

        Example:
            >>> import pysrt
            >>> import codecs
            >>> file = codecs.open('movie.srt', encoding='utf-8')
            >>> for sub in pysrt.stream(file):
            ...     sub.text += "\nHello !"
            ...     print unicode(sub)
        """
        string_buffer = []
        for index, line in enumerate(chain(source_file, '\n')):
            if line.strip():
                string_buffer.append(line)
            else:
                source = string_buffer
                string_buffer = []
                if source and all(source):
                    try:
                        yield SubRipItem.from_lines(source)
                    except Error as error:
                        error.args += (''.join(source), )
                        cls._handle_error(error, error_handling, index)

    def save(self, path=None, encoding=None, eol=None):
        """
        save([path][, encoding][, eol])

        Use initial path if no other provided.
        Use initial encoding if no other provided.
        Use initial eol if no other provided.
        """
        path = path or self.path
        encoding = encoding or self.encoding

        save_file = codecs.open(path, 'w+', encoding=encoding)
        self.write_into(save_file, eol=eol)
        save_file.close()

    def write_into(self, output_file, eol=None):
        """
        write_into(output_file [, eol])

        Serialize current state into `output_file`.

        `output_file` -> Any instance that respond to `write()`, typically a
        file object
        """
        output_eol = eol or self.eol

        for item in self:
            string_repr = str(item)
            if output_eol != '\n':
                string_repr = string_repr.replace('\n', output_eol)
            output_file.write(string_repr)
            # Only add trailing eol if it's not already present.
            # It was kept in the SubRipItem's text before but it really
            # belongs here. Existing applications might give us subtitles
            # which already contain a trailing eol though.
            if not string_repr.endswith(2 * output_eol):
                output_file.write(output_eol)

    @classmethod
    def _guess_eol(cls, string_iterable):
        first_line = cls._get_first_line(string_iterable)
        for eol in ('\r\n', '\r', '\n'):
            if first_line.endswith(eol):
                return eol
        return os.linesep

    @classmethod
    def _get_first_line(cls, string_iterable):
        if hasattr(string_iterable, 'tell'):
            previous_position = string_iterable.tell()

        try:
            first_line = next(iter(string_iterable))
        except StopIteration:
            return ''
        if hasattr(string_iterable, 'seek'):
            string_iterable.seek(previous_position)

        return first_line

    @classmethod
    def _detect_encoding(cls, path):
        file_descriptor = open(path, 'rb')
        first_chars = file_descriptor.read(BIGGER_BOM)
        file_descriptor.close()

        for bom, encoding in BOMS:
            if first_chars.startswith(bom):
                return encoding

        # TODO: maybe a chardet integration
        return cls.DEFAULT_ENCODING

    @classmethod
    def _open_unicode_file(cls, path, claimed_encoding=None):
        encoding = claimed_encoding or cls._detect_encoding(path)
        source_file = codecs.open(path, 'r', encoding=encoding)

        # get rid of BOM if any
        possible_bom = CODECS_BOMS.get(encoding, None)
        if possible_bom:
            file_bom = source_file.read(len(possible_bom))
            if not file_bom == possible_bom:
                source_file.seek(0)  # if not rewind
        return source_file, encoding

    @classmethod
    def _handle_error(cls, error, error_handling, index):
        if error_handling == cls.ERROR_RAISE:
            error.args = (index, ) + error.args
            raise error
        if error_handling == cls.ERROR_LOG:
            name = type(error).__name__
            sys.stderr.write('PySRT-%s(line %s): \n' % (name, index))
            sys.stderr.write(error.args[0].encode('ascii', 'replace'))
            sys.stderr.write('\n')



import os
import shutil
import tarfile
import traceback

from sickchill import logger, settings
from sickchill.oldbeard import helpers, notifiers

from .abstract import UpdateManagerBase


class SourceUpdateManager(UpdateManagerBase):
    def __init__(self):

        self.branch = settings.BRANCH
        if settings.BRANCH == "":
            self.branch = self._find_installed_branch()

        self._cur_commit_hash = settings.CUR_COMMIT_HASH
        self._newest_commit_hash = None
        self._num_commits_behind = 0

        self.session = helpers.make_session()

    @staticmethod
    def _find_installed_branch():
        return settings.CUR_COMMIT_BRANCH if settings.CUR_COMMIT_BRANCH else "master"

    def get_current_commit_hash(self):
        return self._cur_commit_hash

    def get_newest_commit_hash(self):
        return self._newest_commit_hash

    def get_current_version(self):
        return ""

    def get_newest_version(self):
        return ""

    def get_num_commits_behind(self):
        return self._num_commits_behind

    def need_update(self):
        # need this to run first to set self._newest_commit_hash
        try:
            self._check_github_for_update()
        except Exception as e:
            logger.warning("Unable to contact github, can't check for update: " + repr(e))
            return False

        if self.branch != self._find_installed_branch():
            logger.debug("Branch checkout: " + self._find_installed_branch() + "->" + self.branch)
            return True

        if not self._cur_commit_hash or self._num_commits_behind > 0:
            return True

        return False

    def _check_github_for_update(self):
        """
        Uses pygithub to ask github if there is a newer version that the provided
        commit hash. If there is a newer version it sets SickChill's version text.

        commit_hash: hash that we're checking against
        """

        self._num_commits_behind = 0
        self._newest_commit_hash = None

        repo = settings.gh.get_organization(settings.GIT_ORG).get_repo(settings.GIT_REPO)
        # try to get newest commit hash and commits behind directly by comparing branch and current commit
        if self._cur_commit_hash:
            try:
                branch_compared = repo.compare(base=self.branch, head=self._cur_commit_hash)
                self._newest_commit_hash = branch_compared.base_commit.sha
                self._num_commits_behind = branch_compared.behind_by
            except Exception:  # UnknownObjectException
                self._newest_commit_hash = ""
                self._num_commits_behind = 0
                self._cur_commit_hash = ""

        # fall back and iterate over last 100 (items per page in gh_api) commits
        if not self._newest_commit_hash:

            for curCommit in repo.get_commits():
                if not self._newest_commit_hash:
                    self._newest_commit_hash = curCommit.sha
                    if not self._cur_commit_hash:
                        break

                if curCommit.sha == self._cur_commit_hash:
                    break

                # when _cur_commit_hash doesn't match anything _num_commits_behind == 100
                self._num_commits_behind += 1

        logger.debug(
            "cur_commit = {0}, newest_commit = {1}, num_commits_behind = {2}".format(self._cur_commit_hash, self._newest_commit_hash, self._num_commits_behind)
        )

    def set_newest_text(self):
        if not self._cur_commit_hash:
            logger.debug("Unknown current version number, don't know if we should update or not")

            update_url = self.get_update_url()
            newest_tag = "unknown_current_version"
            newest_text = _(
                'Unknown current version number: If you\'ve never used the SickChill upgrade system before then current version is not set. &mdash; <a href="{update_url}">Update Now</a>'.format(
                    update_url=update_url
                )
            )

        elif self._num_commits_behind > 0:
            if self._newest_commit_hash:
                current = self._cur_commit_hash
                newest = self._newest_commit_hash
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/compare/{current}...{newest}"
            else:
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/commits/"

            newest_tag = "newer_version_available"
            commits_behind = self._num_commits_behind
            s = ("", "s")[commits_behind != 1]
            update_url = self.get_update_url()
            newest_text = _(
                'There is a <a href="{url}" onclick="window.open(this.href); return false;">newer version available</a> (you\'re {commits_behind} commit{s} behind) &mdash; <a href="{update_url}">Update Now</a>'.format(
                    commits_behind=commits_behind, update_url=update_url, url=url, s=s
                )
            )
        else:
            return

        helpers.add_site_message(newest_text, tag=newest_tag, level="success")

    def update(self):
        """
        Downloads the latest source tarball from github and installs it over the existing version.
        """

        tar_download_url = "https://github.com/" + settings.GIT_ORG + "/" + settings.GIT_REPO + "/tarball/" + self.branch

        try:
            # prepare the update dir
            sr_update_dir = os.path.join(settings.DATA_DIR, "sr-update")

            if os.path.isdir(sr_update_dir):
                logger.info("Clearing out update folder " + sr_update_dir + " before extracting")
                shutil.rmtree(sr_update_dir)

            logger.info("Creating update folder " + sr_update_dir + " before extracting")
            os.makedirs(sr_update_dir)

            # retrieve file
            logger.info("Downloading update from {url}".format(url=tar_download_url))
            tar_download_path = os.path.join(sr_update_dir, "sr-update.tar")
            helpers.download_file(tar_download_url, tar_download_path, session=self.session)

            if not os.path.isfile(tar_download_path):
                logger.warning("Unable to retrieve new version from " + tar_download_url + ", can't update")
                return False

            if not tarfile.is_tarfile(tar_download_path):
                logger.exception("Retrieved version from " + tar_download_url + " is corrupt, can't update")
                return False

            # extract to sr-update dir
            logger.info("Extracting file " + tar_download_path)
            tar = tarfile.open(tar_download_path)
            tar.extractall(sr_update_dir)
            tar.close()

            # delete .tar.gz
            logger.info("Deleting file " + tar_download_path)
            os.remove(tar_download_path)

            # find update dir name
            update_dir_contents = [x for x in os.listdir(sr_update_dir) if os.path.isdir(os.path.join(sr_update_dir, x))]

            if len(update_dir_contents) != 1:
                logger.exception("Invalid update data, update failed: " + str(update_dir_contents))
                return False

            # walk temp folder and move files to main folder
            content_dir = os.path.join(sr_update_dir, update_dir_contents[0])
            logger.info("Moving files from " + content_dir + " to " + os.path.dirname(settings.PROG_DIR))
            for dirname, stderr_, filenames in os.walk(content_dir):
                dirname = dirname[len(content_dir) + 1 :]
                for curfile in filenames:
                    old_path = os.path.join(content_dir, dirname, curfile)
                    new_path = os.path.join(os.path.dirname(settings.PROG_DIR), dirname, curfile)

                    if os.path.isfile(new_path):
                        os.remove(new_path)
                    os.renames(old_path, new_path)

            settings.CUR_COMMIT_HASH = self._newest_commit_hash
            settings.CUR_COMMIT_BRANCH = self.branch

        except Exception as error:
            logger.exception("Error while trying to update: {}".format(error))
            logger.debug("Traceback: {}".format(traceback.format_exc()))
            return False

        self._clean_libs()

        # Notify update successful
        notifiers.notify_git_update(settings.CUR_COMMIT_HASH or "")
        return True

    def _clean_libs(self):
        lib_path = os.path.join(os.path.dirname(settings.PROG_DIR), "lib3")

        def remove_empty_folders(path):
            if not os.path.isdir(path):
                return

            files = os.listdir(path)
            for f in files:
                full_path = os.path.join(path, f)
                if os.path.isdir(full_path):
                    remove_empty_folders(full_path)

            files = os.listdir(path)
            if len(files) == 0 and path != lib_path:
                os.rmdir(path)

        self._clean_pyc("lib3")
        remove_empty_folders(lib_path)

    def list_remote_branches(self):
        if not settings.gh:
            return []

        repo = settings.gh.get_organization(settings.GIT_ORG).get_repo(settings.GIT_REPO)
        return [x.name for x in repo.get_branches() if x]



import re
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("TorrentLeech")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # URLs
        self.url = "https://www.torrentleech.org"
        self.urls = {
            "login": urljoin(self.url, "user/account/login/"),
            "search": urljoin(self.url, "torrents/browse/list/"),
            "download": urljoin(self.url, "download/"),
        }

        # Proper Strings
        self.proper_strings = ["PROPER", "REPACK"]

        # Cache
        self.cache = tvcache.TVCache(self)

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {
            "username": self.username,
            "password": self.password,
        }

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Invalid Username/password", response) or re.search("<title>Login :: TorrentLeech.org</title>", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        # TV, Episodes, BoxSets, Episodes HD, Animation, Anime, Cartoons
        # 2,26,27,32,7,34,35

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug("Search string: {0}".format(search_string))

                    categories = ["2", "7", "35"]
                    categories += ["26", "32"] if mode == "Episode" else ["27"]
                    if self.show and self.show.is_anime:
                        categories += ["34"]
                else:
                    categories = ["2", "26", "27", "32", "7", "34", "35"]

                # Craft the query URL
                categories_url = "categories/{categories}/".format(categories=",".join(categories))
                query_url = "query/{query_string}".format(query_string=search_string)
                params_url = urljoin(categories_url, query_url)
                search_url = urljoin(self.urls["search"], params_url)

                data = self.get_url(search_url, returns="json")
                if not data:
                    logger.debug("No data returned from provider")
                    continue

                # TODO: Handle more than 35 torrents in return. (Max 35 per call)
                torrent_list = data["torrentList"]

                if len(torrent_list) < 1:
                    logger.debug("Data returned from provider does not contain any torrents")
                    continue

                for torrent in torrent_list:
                    try:
                        title = torrent["name"]
                        download_url = urljoin(self.urls["download"], "{id}/{filename}".format(id=torrent["fid"], filename=torrent["filename"]))

                        seeders = torrent["seeders"]
                        leechers = torrent["leechers"]

                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the"
                                    " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                                )
                                continue

                        size = torrent["size"]

                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}

                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        items.append(item)
                    except Exception:
                        continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["xusto agora", "daquí a un pouco"],
    ["hai %s segundos", "en %s segundos"],
    ["hai 1 minuto", "nun minuto"],
    ["hai %s minutos", "en %s minutos"],
    ["hai 1 hora", "nunha hora"],
    ["hai %s horas", "en %s horas"],
    ["hai 1 día", "nun día"],
    ["hai %s días", "en %s días"],
    ["hai 1 semana", "nunha semana"],
    ["hai %s semanas", "en %s semanas"],
    ["hai 1 mes", "nun mes"],
    ["hai %s meses", "en %s meses"],
    ["hai 1 ano", "nun ano"],
    ["hai %s anos", "en %s anos"]
]
        


import os

from tornado.web import addslash

import sickchill.start
from sickchill import logger, settings
from sickchill.helper import try_int
from sickchill.oldbeard import config, filters, ui
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from . import Config


@Route("/config/notifications(/?.*)", name="config:notifications")
class ConfigNotifications(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @addslash
    def index(self, *args_, **kwargs_):
        t = PageTemplate(rh=self, filename="config_notifications.mako")

        return t.render(
            submenu=self.ConfigMenu(),
            title=_("Config - Notifications"),
            header=_("Notifications"),
            topmenu="config",
            controller="config",
            action="notifications",
        )

    def saveNotifications(
        self,
        use_kodi=None,
        kodi_always_on=None,
        kodi_notify_onsnatch=None,
        kodi_notify_ondownload=None,
        kodi_notify_onsubtitledownload=None,
        kodi_update_onlyfirst=None,
        kodi_update_library=None,
        kodi_update_full=None,
        kodi_host=None,
        kodi_username=None,
        kodi_password=None,
        use_plex_server=None,
        plex_notify_onsnatch=None,
        plex_notify_ondownload=None,
        plex_notify_onsubtitledownload=None,
        plex_update_library=None,
        plex_server_host=None,
        plex_server_token=None,
        plex_client_host=None,
        plex_server_username=None,
        plex_server_password=None,
        use_plex_client=None,
        plex_client_username=None,
        plex_client_password=None,
        plex_server_https=None,
        use_emby=None,
        emby_host=None,
        emby_apikey=None,
        use_growl=None,
        growl_notify_onsnatch=None,
        growl_notify_ondownload=None,
        growl_notify_onsubtitledownload=None,
        growl_host=None,
        growl_password=None,
        use_freemobile=None,
        freemobile_notify_onsnatch=None,
        freemobile_notify_ondownload=None,
        freemobile_notify_onsubtitledownload=None,
        freemobile_id=None,
        freemobile_apikey=None,
        use_telegram=None,
        telegram_notify_onsnatch=None,
        telegram_notify_ondownload=None,
        telegram_notify_onsubtitledownload=None,
        telegram_id=None,
        telegram_apikey=None,
        use_join=None,
        join_notify_onsnatch=None,
        join_notify_ondownload=None,
        join_notify_onsubtitledownload=None,
        join_id=None,
        join_apikey=None,
        use_prowl=None,
        prowl_notify_onsnatch=None,
        prowl_notify_ondownload=None,
        prowl_notify_onsubtitledownload=None,
        prowl_api=None,
        prowl_priority=0,
        prowl_show_list=None,
        prowl_show=None,
        prowl_message_title=None,
        use_twitter=None,
        twitter_notify_onsnatch=None,
        twitter_notify_ondownload=None,
        twitter_notify_onsubtitledownload=None,
        twitter_usedm=None,
        twitter_dmto=None,
        use_twilio=None,
        twilio_notify_onsnatch=None,
        twilio_notify_ondownload=None,
        twilio_notify_onsubtitledownload=None,
        twilio_phone_sid=None,
        twilio_account_sid=None,
        twilio_auth_token=None,
        twilio_to_number=None,
        use_boxcar2=None,
        boxcar2_notify_onsnatch=None,
        boxcar2_notify_ondownload=None,
        boxcar2_notify_onsubtitledownload=None,
        boxcar2_accesstoken=None,
        use_pushover=None,
        pushover_notify_onsnatch=None,
        pushover_notify_ondownload=None,
        pushover_notify_onsubtitledownload=None,
        pushover_userkey=None,
        pushover_apikey=None,
        pushover_device=None,
        pushover_sound=None,
        pushover_priority=0,
        use_libnotify=None,
        libnotify_notify_onsnatch=None,
        libnotify_notify_ondownload=None,
        libnotify_notify_onsubtitledownload=None,
        use_nmj=None,
        nmj_host=None,
        nmj_database=None,
        nmj_mount=None,
        use_synoindex=None,
        use_nmjv2=None,
        nmjv2_host=None,
        nmjv2_dbloc=None,
        nmjv2_database=None,
        use_trakt=None,
        trakt_username=None,
        trakt_pin=None,
        trakt_remove_watchlist=None,
        trakt_sync_watchlist=None,
        trakt_remove_show_from_sickchill=None,
        trakt_method_add=None,
        trakt_start_paused=None,
        trakt_use_recommended=None,
        trakt_sync=None,
        trakt_sync_remove=None,
        trakt_default_indexer=None,
        trakt_remove_serieslist=None,
        trakt_timeout=None,
        trakt_blacklist_name=None,
        use_synologynotifier=None,
        synologynotifier_notify_onsnatch=None,
        synologynotifier_notify_ondownload=None,
        synologynotifier_notify_onsubtitledownload=None,
        use_pytivo=None,
        pytivo_notify_onsnatch=None,
        pytivo_notify_ondownload=None,
        pytivo_notify_onsubtitledownload=None,
        pytivo_update_library=None,
        pytivo_host=None,
        pytivo_share_name=None,
        pytivo_tivo_name=None,
        use_pushalot=None,
        pushalot_notify_onsnatch=None,
        pushalot_notify_ondownload=None,
        pushalot_notify_onsubtitledownload=None,
        pushalot_authorizationtoken=None,
        use_pushbullet=None,
        pushbullet_notify_onsnatch=None,
        pushbullet_notify_ondownload=None,
        pushbullet_notify_onsubtitledownload=None,
        pushbullet_api=None,
        pushbullet_device=None,
        pushbullet_device_list=None,
        pushbullet_channel_list=None,
        pushbullet_channel=None,
        use_email=None,
        email_notify_onsnatch=None,
        email_notify_ondownload=None,
        email_notify_onpostprocess=None,
        email_notify_onsubtitledownload=None,
        email_host=None,
        email_port=25,
        email_from=None,
        email_tls=None,
        email_user=None,
        email_password=None,
        email_list=None,
        email_subject=None,
        email_show_list=None,
        email_show=None,
        use_slack=False,
        slack_notify_snatch=None,
        slack_notify_download=None,
        slack_notify_subtitledownload=None,
        slack_webhook=None,
        slack_icon_emoji=None,
        use_rocketchat=False,
        rocketchat_notify_snatch=None,
        rocketchat_notify_download=None,
        rocketchat_notify_subtitledownload=None,
        rocketchat_webhook=None,
        rocketchat_icon_emoji=None,
        use_matrix=False,
        matrix_notify_snatch=None,
        matrix_notify_download=None,
        matrix_notify_subtitledownload=None,
        matrix_api_token=None,
        matrix_server=None,
        matrix_room=None,
        use_discord=False,
        discord_notify_snatch=None,
        discord_notify_download=None,
        discord_webhook=None,
        discord_name=None,
        discord_avatar_url=None,
        discord_tts=False,
    ):

        results = []

        settings.USE_KODI = config.checkbox_to_value(use_kodi)
        settings.KODI_ALWAYS_ON = config.checkbox_to_value(kodi_always_on)
        settings.KODI_NOTIFY_ONSNATCH = config.checkbox_to_value(kodi_notify_onsnatch)
        settings.KODI_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(kodi_notify_ondownload)
        settings.KODI_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(kodi_notify_onsubtitledownload)
        settings.KODI_UPDATE_LIBRARY = config.checkbox_to_value(kodi_update_library)
        settings.KODI_UPDATE_FULL = config.checkbox_to_value(kodi_update_full)
        settings.KODI_UPDATE_ONLYFIRST = config.checkbox_to_value(kodi_update_onlyfirst)
        settings.KODI_HOST = config.clean_hosts(kodi_host)
        settings.KODI_USERNAME = kodi_username
        settings.KODI_PASSWORD = filters.unhide(settings.KODI_PASSWORD, kodi_password)

        settings.USE_PLEX_SERVER = config.checkbox_to_value(use_plex_server)
        settings.PLEX_NOTIFY_ONSNATCH = config.checkbox_to_value(plex_notify_onsnatch)
        settings.PLEX_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(plex_notify_ondownload)
        settings.PLEX_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(plex_notify_onsubtitledownload)
        settings.PLEX_UPDATE_LIBRARY = config.checkbox_to_value(plex_update_library)
        settings.PLEX_CLIENT_HOST = config.clean_hosts(plex_client_host)
        settings.PLEX_SERVER_HOST = config.clean_hosts(plex_server_host)
        settings.PLEX_SERVER_TOKEN = config.clean_host(plex_server_token)
        settings.PLEX_SERVER_USERNAME = plex_server_username
        settings.PLEX_SERVER_PASSWORD = filters.unhide(settings.PLEX_SERVER_PASSWORD, plex_server_password)

        settings.USE_PLEX_CLIENT = config.checkbox_to_value(use_plex_client)
        settings.PLEX_CLIENT_USERNAME = plex_client_username
        settings.PLEX_CLIENT_PASSWORD = filters.unhide(settings.PLEX_CLIENT_PASSWORD, plex_client_password)
        settings.PLEX_SERVER_HTTPS = config.checkbox_to_value(plex_server_https)

        settings.USE_EMBY = config.checkbox_to_value(use_emby)
        settings.EMBY_HOST = config.clean_url(emby_host)
        settings.EMBY_APIKEY = filters.unhide(settings.EMBY_APIKEY, emby_apikey)

        settings.USE_GROWL = config.checkbox_to_value(use_growl)
        settings.GROWL_NOTIFY_ONSNATCH = config.checkbox_to_value(growl_notify_onsnatch)
        settings.GROWL_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(growl_notify_ondownload)
        settings.GROWL_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(growl_notify_onsubtitledownload)
        settings.GROWL_HOST = config.clean_host(growl_host, default_port=23053)
        settings.GROWL_PASSWORD = filters.unhide(settings.GROWL_PASSWORD, growl_password)

        settings.USE_FREEMOBILE = config.checkbox_to_value(use_freemobile)
        settings.FREEMOBILE_NOTIFY_ONSNATCH = config.checkbox_to_value(freemobile_notify_onsnatch)
        settings.FREEMOBILE_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(freemobile_notify_ondownload)
        settings.FREEMOBILE_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(freemobile_notify_onsubtitledownload)
        settings.FREEMOBILE_ID = freemobile_id
        settings.FREEMOBILE_APIKEY = filters.unhide(settings.FREEMOBILE_APIKEY, freemobile_apikey)

        settings.USE_TELEGRAM = config.checkbox_to_value(use_telegram)
        settings.TELEGRAM_NOTIFY_ONSNATCH = config.checkbox_to_value(telegram_notify_onsnatch)
        settings.TELEGRAM_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(telegram_notify_ondownload)
        settings.TELEGRAM_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(telegram_notify_onsubtitledownload)
        settings.TELEGRAM_ID = telegram_id
        settings.TELEGRAM_APIKEY = filters.unhide(settings.TELEGRAM_APIKEY, telegram_apikey)

        settings.USE_JOIN = config.checkbox_to_value(use_join)
        settings.JOIN_NOTIFY_ONSNATCH = config.checkbox_to_value(join_notify_onsnatch)
        settings.JOIN_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(join_notify_ondownload)
        settings.JOIN_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(join_notify_onsubtitledownload)
        settings.JOIN_ID = join_id
        settings.JOIN_APIKEY = filters.unhide(settings.JOIN_APIKEY, join_apikey)

        settings.USE_PROWL = config.checkbox_to_value(use_prowl)
        settings.PROWL_NOTIFY_ONSNATCH = config.checkbox_to_value(prowl_notify_onsnatch)
        settings.PROWL_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(prowl_notify_ondownload)
        settings.PROWL_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(prowl_notify_onsubtitledownload)
        settings.PROWL_API = prowl_api
        settings.PROWL_PRIORITY = prowl_priority
        settings.PROWL_MESSAGE_TITLE = prowl_message_title

        settings.USE_TWITTER = config.checkbox_to_value(use_twitter)
        settings.TWITTER_NOTIFY_ONSNATCH = config.checkbox_to_value(twitter_notify_onsnatch)
        settings.TWITTER_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(twitter_notify_ondownload)
        settings.TWITTER_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(twitter_notify_onsubtitledownload)
        settings.TWITTER_USEDM = config.checkbox_to_value(twitter_usedm)
        settings.TWITTER_DMTO = twitter_dmto

        settings.USE_TWILIO = config.checkbox_to_value(use_twilio)
        settings.TWILIO_NOTIFY_ONSNATCH = config.checkbox_to_value(twilio_notify_onsnatch)
        settings.TWILIO_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(twilio_notify_ondownload)
        settings.TWILIO_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(twilio_notify_onsubtitledownload)
        settings.TWILIO_PHONE_SID = twilio_phone_sid
        settings.TWILIO_ACCOUNT_SID = twilio_account_sid
        settings.TWILIO_AUTH_TOKEN = twilio_auth_token
        settings.TWILIO_TO_NUMBER = twilio_to_number

        settings.USE_SLACK = config.checkbox_to_value(use_slack)
        settings.SLACK_NOTIFY_SNATCH = config.checkbox_to_value(slack_notify_snatch)
        settings.SLACK_NOTIFY_DOWNLOAD = config.checkbox_to_value(slack_notify_download)
        settings.SLACK_NOTIFY_SUBTITLEDOWNLOAD = config.checkbox_to_value(slack_notify_subtitledownload)
        settings.SLACK_WEBHOOK = slack_webhook
        settings.SLACK_ICON_EMOJI = slack_icon_emoji

        settings.USE_ROCKETCHAT = config.checkbox_to_value(use_rocketchat)
        settings.ROCKETCHAT_NOTIFY_SNATCH = config.checkbox_to_value(rocketchat_notify_snatch)
        settings.ROCKETCHAT_NOTIFY_DOWNLOAD = config.checkbox_to_value(rocketchat_notify_download)
        settings.ROCKETCHAT_NOTIFY_SUBTITLEDOWNLOAD = config.checkbox_to_value(rocketchat_notify_subtitledownload)
        settings.ROCKETCHAT_WEBHOOK = rocketchat_webhook
        settings.ROCKETCHAT_ICON_EMOJI = rocketchat_icon_emoji

        settings.USE_MATRIX = config.checkbox_to_value(use_matrix)
        settings.MATRIX_NOTIFY_SNATCH = config.checkbox_to_value(matrix_notify_snatch)
        settings.MATRIX_NOTIFY_DOWNLOAD = config.checkbox_to_value(matrix_notify_download)
        settings.MATRIX_NOTIFY_SUBTITLEDOWNLOAD = config.checkbox_to_value(matrix_notify_subtitledownload)
        settings.MATRIX_API_TOKEN = matrix_api_token
        settings.MATRIX_SERVER = matrix_server
        settings.MATRIX_ROOM = matrix_room

        settings.USE_DISCORD = config.checkbox_to_value(use_discord)
        settings.DISCORD_NOTIFY_SNATCH = config.checkbox_to_value(discord_notify_snatch)
        settings.DISCORD_NOTIFY_DOWNLOAD = config.checkbox_to_value(discord_notify_download)
        settings.DISCORD_WEBHOOK = discord_webhook
        settings.DISCORD_NAME = discord_name
        settings.DISCORD_AVATAR_URL = discord_avatar_url
        settings.DISCORD_TTS = discord_tts

        settings.USE_BOXCAR2 = config.checkbox_to_value(use_boxcar2)
        settings.BOXCAR2_NOTIFY_ONSNATCH = config.checkbox_to_value(boxcar2_notify_onsnatch)
        settings.BOXCAR2_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(boxcar2_notify_ondownload)
        settings.BOXCAR2_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(boxcar2_notify_onsubtitledownload)
        settings.BOXCAR2_ACCESSTOKEN = boxcar2_accesstoken

        settings.USE_PUSHOVER = config.checkbox_to_value(use_pushover)
        settings.PUSHOVER_NOTIFY_ONSNATCH = config.checkbox_to_value(pushover_notify_onsnatch)
        settings.PUSHOVER_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(pushover_notify_ondownload)
        settings.PUSHOVER_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(pushover_notify_onsubtitledownload)
        settings.PUSHOVER_USERKEY = pushover_userkey
        settings.PUSHOVER_APIKEY = filters.unhide(settings.PUSHOVER_APIKEY, pushover_apikey)
        settings.PUSHOVER_DEVICE = pushover_device
        settings.PUSHOVER_SOUND = pushover_sound
        settings.PUSHOVER_PRIORITY = pushover_priority

        settings.USE_LIBNOTIFY = config.checkbox_to_value(use_libnotify)
        settings.LIBNOTIFY_NOTIFY_ONSNATCH = config.checkbox_to_value(libnotify_notify_onsnatch)
        settings.LIBNOTIFY_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(libnotify_notify_ondownload)
        settings.LIBNOTIFY_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(libnotify_notify_onsubtitledownload)

        settings.USE_NMJ = config.checkbox_to_value(use_nmj)
        settings.NMJ_HOST = config.clean_host(nmj_host)
        settings.NMJ_DATABASE = nmj_database
        settings.NMJ_MOUNT = nmj_mount

        settings.USE_NMJv2 = config.checkbox_to_value(use_nmjv2)
        settings.NMJv2_HOST = config.clean_host(nmjv2_host)
        settings.NMJv2_DATABASE = nmjv2_database
        settings.NMJv2_DBLOC = nmjv2_dbloc

        settings.USE_SYNOINDEX = config.checkbox_to_value(use_synoindex)

        settings.USE_SYNOLOGYNOTIFIER = config.checkbox_to_value(use_synologynotifier)
        settings.SYNOLOGYNOTIFIER_NOTIFY_ONSNATCH = config.checkbox_to_value(synologynotifier_notify_onsnatch)
        settings.SYNOLOGYNOTIFIER_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(synologynotifier_notify_ondownload)
        settings.SYNOLOGYNOTIFIER_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(synologynotifier_notify_onsubtitledownload)

        config.change_use_trakt(use_trakt)
        settings.TRAKT_USERNAME = trakt_username
        settings.TRAKT_REMOVE_WATCHLIST = config.checkbox_to_value(trakt_remove_watchlist)
        settings.TRAKT_REMOVE_SERIESLIST = config.checkbox_to_value(trakt_remove_serieslist)
        settings.TRAKT_REMOVE_SHOW_FROM_SICKCHILL = config.checkbox_to_value(trakt_remove_show_from_sickchill)
        settings.TRAKT_SYNC_WATCHLIST = config.checkbox_to_value(trakt_sync_watchlist)
        settings.TRAKT_METHOD_ADD = int(trakt_method_add)
        settings.TRAKT_START_PAUSED = config.checkbox_to_value(trakt_start_paused)
        settings.TRAKT_USE_RECOMMENDED = config.checkbox_to_value(trakt_use_recommended)
        settings.TRAKT_SYNC = config.checkbox_to_value(trakt_sync)
        settings.TRAKT_SYNC_REMOVE = config.checkbox_to_value(trakt_sync_remove)
        settings.TRAKT_DEFAULT_INDEXER = int(trakt_default_indexer)
        settings.TRAKT_TIMEOUT = int(trakt_timeout)
        settings.TRAKT_BLACKLIST_NAME = trakt_blacklist_name

        settings.USE_EMAIL = config.checkbox_to_value(use_email)
        settings.EMAIL_NOTIFY_ONSNATCH = config.checkbox_to_value(email_notify_onsnatch)
        settings.EMAIL_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(email_notify_ondownload)
        settings.EMAIL_NOTIFY_ONPOSTPROCESS = config.checkbox_to_value(email_notify_onpostprocess)
        settings.EMAIL_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(email_notify_onsubtitledownload)
        settings.EMAIL_HOST = config.clean_host(email_host)
        settings.EMAIL_PORT = try_int(email_port, 25)
        settings.EMAIL_FROM = email_from
        settings.EMAIL_TLS = config.checkbox_to_value(email_tls)
        settings.EMAIL_USER = email_user
        settings.EMAIL_PASSWORD = filters.unhide(settings.EMAIL_PASSWORD, email_password)
        settings.EMAIL_LIST = email_list
        settings.EMAIL_SUBJECT = email_subject

        settings.USE_PYTIVO = config.checkbox_to_value(use_pytivo)
        settings.PYTIVO_NOTIFY_ONSNATCH = config.checkbox_to_value(pytivo_notify_onsnatch)
        settings.PYTIVO_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(pytivo_notify_ondownload)
        settings.PYTIVO_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(pytivo_notify_onsubtitledownload)
        settings.PYTIVO_UPDATE_LIBRARY = config.checkbox_to_value(pytivo_update_library)
        settings.PYTIVO_HOST = config.clean_host(pytivo_host)
        settings.PYTIVO_SHARE_NAME = pytivo_share_name
        settings.PYTIVO_TIVO_NAME = pytivo_tivo_name

        settings.USE_PUSHALOT = config.checkbox_to_value(use_pushalot)
        settings.PUSHALOT_NOTIFY_ONSNATCH = config.checkbox_to_value(pushalot_notify_onsnatch)
        settings.PUSHALOT_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(pushalot_notify_ondownload)
        settings.PUSHALOT_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(pushalot_notify_onsubtitledownload)
        settings.PUSHALOT_AUTHORIZATIONTOKEN = pushalot_authorizationtoken

        settings.USE_PUSHBULLET = config.checkbox_to_value(use_pushbullet)
        settings.PUSHBULLET_NOTIFY_ONSNATCH = config.checkbox_to_value(pushbullet_notify_onsnatch)
        settings.PUSHBULLET_NOTIFY_ONDOWNLOAD = config.checkbox_to_value(pushbullet_notify_ondownload)
        settings.PUSHBULLET_NOTIFY_ONSUBTITLEDOWNLOAD = config.checkbox_to_value(pushbullet_notify_onsubtitledownload)
        settings.PUSHBULLET_API = pushbullet_api
        settings.PUSHBULLET_DEVICE = pushbullet_device_list
        settings.PUSHBULLET_CHANNEL = pushbullet_channel_list or ""

        sickchill.start.save_config()

        if len(results) > 0:
            for x in results:
                logger.exception(x)
            ui.notifications.error(_("Error(s) Saving Configuration"), "<br>\n".join(results))
        else:
            ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/notifications/")



# -*- coding: utf-8 -*-
from rebulk.loose import ensure_list

from .score import get_equivalent_release_groups, score_keys
from .video import Episode, Movie
from .utils import sanitize, sanitize_release_group


def series_matches(video, title=None, **kwargs):
    """Whether the `video` matches the series title.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str title: the series name.
    :return: whether there's a match
    :rtype: bool

    """
    if isinstance(video, Episode):
        return video.series and sanitize(title) in (
            sanitize(name) for name in [video.series] + video.alternative_series
        )


def title_matches(video, title=None, episode_title=None, **kwargs):
    """Whether the movie matches the movie `title` or the series matches the `episode_title`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str title: the movie title.
    :param str episode_title: the series episode title.
    :return: whether there's a match
    :rtype: bool

    """
    if isinstance(video, Episode):
        return video.title and sanitize(episode_title) == sanitize(video.title)
    if isinstance(video, Movie):
        return video.title and sanitize(title) == sanitize(video.title)


def season_matches(video, season=None, **kwargs):
    """Whether the episode matches the `season`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param int season: the episode season.
    :return: whether there's a match
    :rtype: bool

    """
    if isinstance(video, Episode):
        return video.season and season == video.season


def episode_matches(video, episode=None, **kwargs):
    """Whether the episode matches the `episode`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param episode: the episode season.
    :type: list of int or int
    :return: whether there's a match
    :rtype: bool

    """
    if isinstance(video, Episode):
        return video.episodes and ensure_list(episode) == video.episodes


def year_matches(video, year=None, partial=False, **kwargs):
    """Whether the video matches the `year`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param int year: the video year.
    :param bool partial: whether or not the guess is partial.
    :return: whether there's a match
    :rtype: bool

    """
    if video.year and year == video.year:
        return True
    if isinstance(video, Episode):
        # count "no year" as an information
        return not partial and video.original_series and not year


def country_matches(video, country=None, partial=False, **kwargs):
    """Whether the video matches the `country`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param country: the video country.
    :type country: :class:`~babelfish.country.Country`
    :param bool partial: whether or not the guess is partial.
    :return: whether there's a match
    :rtype: bool

    """
    if video.country and country == video.country:
        return True

    if isinstance(video, Episode):
        # count "no country" as an information
        return not partial and video.original_series and not country

    if isinstance(video, Movie):
        # count "no country" as an information
        return not video.country and not country


def release_group_matches(video, release_group=None, **kwargs):
    """Whether the video matches the `release_group`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str release_group: the video release group.
    :return: whether there's a match
    :rtype: bool

    """
    return (video.release_group and release_group and
            any(r in sanitize_release_group(release_group)
                for r in get_equivalent_release_groups(sanitize_release_group(video.release_group))))


def streaming_service_matches(video, streaming_service=None, **kwargs):
    """Whether the video matches the `streaming_service`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str streaming_service: the video streaming service
    :return: whether there's a match
    :rtype: bool

    """
    return video.streaming_service and streaming_service == video.streaming_service


def resolution_matches(video, screen_size=None, **kwargs):
    """Whether the video matches the `resolution`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str screen_size: the video resolution
    :return: whether there's a match
    :rtype: bool

    """
    return video.resolution and screen_size == video.resolution


def source_matches(video, source=None, **kwargs):
    """Whether the video matches the `source`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str source: the video source
    :return: whether there's a match
    :rtype: bool

    """
    return video.source and source == video.source


def video_codec_matches(video, video_codec=None, **kwargs):
    """Whether the video matches the `video_codec`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str video_codec: the video codec
    :return: whether there's a match
    :rtype: bool

    """
    return video.video_codec and video_codec == video.video_codec


def audio_codec_matches(video, audio_codec=None, **kwargs):
    """Whether the video matches the `audio_codec`.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param str audio_codec: the video audio codec
    :return: whether there's a match
    :rtype: bool

    """
    return video.audio_codec and audio_codec == video.audio_codec


#: Available matches functions
matches_manager = {
    'series': series_matches,
    'title': title_matches,
    'season': season_matches,
    'episode': episode_matches,
    'year': year_matches,
    'country': country_matches,
    'release_group': release_group_matches,
    'streaming_service': streaming_service_matches,
    'resolution': resolution_matches,
    'source': source_matches,
    'video_codec': video_codec_matches,
    'audio_codec': audio_codec_matches
}


def guess_matches(video, guess, partial=False):
    """Get matches between a `video` and a `guess`.

    If a guess is `partial`, the absence information won't be counted as a match.

    :param video: the video.
    :type video: :class:`~subliminal.video.Video`
    :param guess: the guess.
    :type guess: dict
    :param bool partial: whether or not the guess is partial.
    :return: matches between the `video` and the `guess`.
    :rtype: set

    """
    matches = set()
    for key in score_keys:
        if key in matches_manager and matches_manager[key](video, partial=partial, **guess):
            matches.add(key)

    return matches



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import os
import io
import json
import logging
import binascii
import webbrowser
import pkg_resources
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from datetime import datetime

import tus
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

__version__ = pkg_resources.get_distribution('putio.py').version

KB = 1024
MB = 1024 * KB

# Read and write operations are limited to this chunk size.
# This can make a big difference when dealing with large files.
CHUNK_SIZE = 256 * KB

BASE_URL = None
UPLOAD_URL = None
TUS_UPLOAD_URL = None
ACCESS_TOKEN_URL = None
AUTHENTICATION_URL = None
AUTHORIZATION_URL = None

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _set_domain(domain='put.io', scheme='https'):
    global BASE_URL
    global UPLOAD_URL
    global TUS_UPLOAD_URL
    global ACCESS_TOKEN_URL
    global AUTHENTICATION_URL
    global AUTHORIZATION_URL

    api_base = '{scheme}://api.{domain}/v2'.format(scheme=scheme, domain=domain)
    upload_base = '{scheme}://upload.{domain}'.format(scheme=scheme, domain=domain)

    BASE_URL = api_base
    UPLOAD_URL = upload_base + '/v2/files/upload'
    TUS_UPLOAD_URL = upload_base + '/files/'
    ACCESS_TOKEN_URL = api_base + '/oauth2/access_token'
    AUTHENTICATION_URL = api_base + '/oauth2/authenticate'
    AUTHORIZATION_URL = api_base + '/oauth2/authorizations/clients/{client_id}/{fingerprint}'


_set_domain()


class APIError(Exception):
    """
    Must be created with following arguments:
        1. Response instance (requests.Response)
        2. Type of the error (str)
        3. Extra detail about the error (str, optional)
    """

    def __str__(self):
        s = "%s, %s, %d, %s" % (
                self.response.request.method,
                self.response.request.url,
                self.response.status_code,
                self.type,
        )
        if self.message:
            s += ', %r' % self.message
        return s

    @property
    def response(self):
        return self.args[0]

    @property
    def type(self):
        return self.args[1]

    @property
    def message(self):
        if len(self.args) > 2:
            return self.args[2]


class ClientError(APIError):
    pass


class ServerError(APIError):
    pass


class AuthHelper(object):

    def __init__(self, client_id, client_secret, redirect_uri, type='code'):
        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_url = redirect_uri
        self.type = type

    @property
    def authentication_url(self):
        """Redirect your users to here to authenticate them."""
        params = {
            'client_id': self.client_id,
            'response_type': self.type,
            'redirect_uri': self.callback_url
        }
        return AUTHENTICATION_URL + "?" + urlencode(params)

    def open_authentication_url(self):
        webbrowser.open(self.authentication_url)

    def get_access_token(self, code):
        params = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'authorization_code',
            'redirect_uri': self.callback_url,
            'code': code
        }
        response = requests.get(ACCESS_TOKEN_URL, params=params)
        return _process_response(response)['access_token']


def create_access_token(client_id, client_secret, user, password, fingerprint=''):
    url = AUTHORIZATION_URL.format(client_id=client_id, fingerprint=fingerprint)
    data = {'client_secret': client_secret}
    auth = (user, password)
    response = requests.put(url, data=data, auth=auth)
    return _process_response(response)['access_token']


def revoke_access_token(access_token):
    url = BASE_URL + '/oauth/grants/logout'
    headers = {'Authorization': 'token %s' % access_token}
    response = requests.post(url, headers=headers)
    _process_response(response)


class Client(object):

    def __init__(self, access_token, use_retry=False, extra_headers=None, timeout=5):
        self.access_token = access_token
        self.session = requests.session()
        self.session.headers['User-Agent'] = 'putio.py/%s' % __version__
        self.session.headers['Accept'] = 'application/json'
        self.timeout = timeout
        if extra_headers:
            self.session.headers.update(extra_headers)

        if use_retry:
            # Retry maximum 10 times, backoff on each retry
            # Sleeps 1s, 2s, 4s, 8s, etc to a maximum of 120s between retries
            # Retries on HTTP status codes 500, 502, 503, 504
            retries = Retry(total=10,
                            backoff_factor=1,
                            status_forcelist=[500, 502, 503, 504])

            # Use the retry strategy for all HTTPS requests
            self.session.mount('https://', HTTPAdapter(max_retries=retries))

        # Keep resource classes as attributes of client.
        # Pass client to resource classes so resource object
        # can use the client.
        attributes = {'client': self}
        self.File = type('File', (_File,), attributes)
        self.Subtitle = type('Subtitle', (_Subtitle,), attributes)
        self.Transfer = type('Transfer', (_Transfer,), attributes)
        self.Account = type('Account', (_Account,), attributes)

    def close(self):
        self.session.close()

    def request(self, path, method='GET', params=None, data=None, files=None,
                headers=None, raw=False, allow_redirects=True, stream=False,
                timeout=None):
        """
        Wrapper around requests.request()

        Prepends BASE_URL to path.
        Adds self.oauth_token to authorization header.
        Parses response as JSON and returns it.

        """
        if not headers:
            headers = {}

        if timeout is None:
            timeout = self.timeout

        # All requests must include oauth_token
        headers['Authorization'] = 'token %s' % self.access_token

        if path.startswith(('http://', 'https://')):
            url = path
        else:
            url = BASE_URL + path
        logger.debug('url: %s', url)

        response = self.session.request(
            method, url, params=params, data=data, files=files,
            headers=headers, allow_redirects=allow_redirects, stream=stream,
            timeout=self.timeout)
        logger.debug('response: %s', response)
        if raw:
            return response

        return _process_response(response)


def _process_response(response):
    logger.debug('response: %s', response)
    logger.debug('content: %s', response.content)

    http_error_type = str(response.status_code)[0]
    exception_classes = {
            '2': None,
            '4': ClientError,
            '5': ServerError,
    }

    try:
        exception_class = exception_classes[http_error_type]
    except KeyError:
        raise ServerError(response, 'UnknownStatusCode', str(response.status_code))

    if exception_class:
        try:
            d = _parse_content(response)
            error_type = d['error_type']
            error_message = d['error_message']
        except Exception:
            error_type = 'UnknownError'
            error_message = None

        raise exception_class(response, error_type, error_message)

    return _parse_content(response)


def _parse_content(response):
    try:
        u = response.content.decode('utf-8')
    except ValueError:
        raise ServerError(response, 'InvalidEncoding', 'cannot decode as UTF-8')

    try:
        return json.loads(u)
    except ValueError:
        raise ServerError(response, 'InvalidJSON')


class _BaseResource(object):

    client = None

    def __init__(self, resource_dict):
        """Constructs the object from a dict."""
        # All resources must have id and name attributes
        self.id = None
        self.name = None
        self.__dict__.update(resource_dict)
        try:
            self.created_at = strptime(self.created_at)
        except Exception:
            self.created_at = None

    def __str__(self):
        return self.name.encode('utf-8')

    def __repr__(self):
        # shorten name for display
        name = self.name[:17] + '...' if len(self.name) > 20 else self.name
        return '<%s id=%r, name=%r>' % (
            self.__class__.__name__, self.id, name)


class _File(_BaseResource):

    @classmethod
    def get(cls, id):
        d = cls.client.request('/files/%i' % id, method='GET')
        t = d['file']
        return cls(t)

    @classmethod
    def list(cls, parent_id=0, per_page=1000, sort_by=None, content_type=None,
             file_type=None, stream_url=False, stream_url_parent=False, mp4_stream_url=False,
             mp4_stream_url_parent=False, hidden=False, mp4_status=False):
        """ List files and their properties.

         parent_id List files under a folder. If not specified, it will show files listed at the root folder

         """
        params = {
                'parent_id': parent_id,
                'per_page': str(per_page),
                'sort_by': sort_by or '',
                'content_type': content_type or '',
                'file_type': file_type or '',
                'stream_url': str(stream_url),
                'stream_url_parent': str(stream_url_parent),
                'mp4_stream_url':  str(mp4_stream_url),
                'mp4_stream_url_parent': str(mp4_stream_url_parent),
                'hidden': str(hidden),
                'mp4_status':  str(mp4_status),
        }
        d = cls.client.request('/files/list', params=params)
        files = d['files']
        while d['cursor']:
            d = cls.client.request('/files/list/continue', method='POST', data={'cursor': d['cursor']})
            files.extend(d['files'])

        return [cls(f) for f in files]

    @classmethod
    def upload(cls, path, name=None, parent_id=0):
        """ If the uploaded file is a torrent file, starts it as a transfer. This endpoint must be used with upload.put.io domain.
        name: override the file name
        parent_id: where to put the file
        """
        with io.open(path, 'rb') as f:
            if name:
                files = {'file': (name, f)}
            else:
                files = {'file': f}
            d = cls.client.request(UPLOAD_URL, method='POST',
                                   data={'parent_id': parent_id}, files=files)

        try:
            return cls(d['file'])
        except KeyError:
            # server returns a transfer info if file is a torrent
            return cls.client.Transfer(d['transfer'])

    @classmethod
    def upload_tus(cls, path, name=None, parent_id=0):
        headers = {'Authorization': 'token %s' % cls.client.access_token}
        metadata = {'parent_id': str(parent_id)}
        if name:
            metadata['name'] = name
        else:
            metadata['name'] = os.path.basename(path)
        with io.open(path, 'rb') as f:
            tus.upload(f, TUS_UPLOAD_URL, file_name=name, headers=headers, metadata=metadata)

    @classmethod
    def search(cls, query, per_page=100):
        """
        Search makes a search request with the given query
        query: The keyword to search
        per_page: Number of files to be returned in response.
        """
        path = '/files/search'
        result = cls.client.request(path, params={'query': query, 'per_page': per_page})
        files = result['files']
        return [cls(f) for f in files]

    def dir(self):
        """List the files under directory."""
        return self.list(parent_id=self.id)

    def download(self, dest='.', delete_after_download=False, chunk_size=CHUNK_SIZE):
        if self.content_type == 'application/x-directory':
            self._download_directory(dest, delete_after_download, chunk_size)
        else:
            self._download_file(dest, delete_after_download, chunk_size)

    def _download_directory(self, dest, delete_after_download, chunk_size):
        name = _str(self.name)

        dest = os.path.join(dest, name)
        if not os.path.exists(dest):
            os.mkdir(dest)

        for sub_file in self.dir():
            sub_file.download(dest, delete_after_download, chunk_size)

        if delete_after_download:
            self.delete()

    def _verify_file(self, filepath):
        logger.info('verifying crc32...')
        filesize = os.path.getsize(filepath)
        if self.size != filesize:
            logging.error('file %s is %d bytes, should be %s bytes' % (filepath, filesize, self.size))
            return False

        crcbin = 0
        with io.open(filepath, 'rb') as f:
            while True:
                chunk = f.read(CHUNK_SIZE)
                if not chunk:
                    break

                crcbin = binascii.crc32(chunk, crcbin) & 0xffffffff

        crc32 = '%08x' % crcbin

        if crc32 != self.crc32:
            logging.error('file %s CRC32 is %s, should be %s' % (filepath, crc32, self.crc32))
            return False

        logger.info('crc OK')
        return True

    def _download_file(self, dest, delete_after_download, chunk_size):
        name = _str(self.name)

        filepath = os.path.join(dest, name)
        logger.info('downloading file to: %s', filepath)
        if os.path.exists(filepath):
            first_byte = os.path.getsize(filepath)

            if first_byte == self.size:
                logger.warning('file %s exists and is the correct size %d' % (filepath, self.size))
        else:
            first_byte = 0

        logger.debug('file %s is currently %d, should be %d' % (filepath, first_byte, self.size))

        if self.size == 0:
            # Create an empty file
            io.open(filepath, 'wb').close()
            logger.debug('created empty file %s' % filepath)
        else:
            if first_byte < self.size:
                with io.open(filepath, 'ab') as f:
                    headers = {'Range': 'bytes=%d-' % first_byte}

                    logger.debug('request range: bytes=%d-' % first_byte)
                    path = '/files/%d/url' % self.id
                    response = self.client.request(path, raw=True)
                    if str(response.status_code)[0] != '2':
                        # Raises exception on 4xx and 5xx
                        _process_response(response)
                    
                    download_link = str(response.json().get('url'))
                    response = self.client.request(download_link,
                                                   headers=headers,
                                                   raw=True,
                                                   stream=True)
                    if str(response.status_code)[0] != '2':
                        # Raises exception on 4xx and 5xx
                        _process_response(response)

                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)

        if self._verify_file(filepath):
            if delete_after_download:
                self.delete()

    def _get_link(self, path, params):
        response = self.client.request(path, method='HEAD', params=params, raw=True, allow_redirects=False)
        if str(response.status_code)[0] == '2':
            return response.url
        elif response.status_code == 302:
            return response.headers['Location']

        # Raises exception on 4xx and 5xx
        _process_response(response)

    def get_stream_link(self, tunnel=True, prefer_mp4=False):
        if prefer_mp4 and self.get_mp4_status()['status'] == 'COMPLETED':
            path = '/files/%d/mp4/stream' % self.id
        else:
            path = '/files/%d/stream' % self.id

        params = {}
        if not tunnel:
            params['notunnel'] = '1'

        return self._get_link(path, params)

    def get_download_link(self):
        path = '/files/%d/download' % self.id
        params = {}

        return self._get_link(path, params)

    def convert_to_mp4(self):
        path = '/files/%d/mp4' % self.id
        self.client.request(path, method='POST')

    def get_mp4_status(self):
        path = '/files/%d/mp4' % self.id
        response = self.client.request(path)
        return response['mp4']

    def get_subtitles(cls):
        path = '/files/%d/subtitles' % cls.id
        response = cls.client.request(path, method='GET')
        json_subtitles = response['subtitles']

        return [cls.client.Subtitle(s, cls.id) for s in json_subtitles]

    def delete(self, skip_nonexistents=False):
        data = {'file_id': self.id}
        if skip_nonexistents:
            data['skip_nonexistents'] = 1
        return self.client.request('/files/delete', method='POST', data=data)

    @classmethod
    def delete_multi(cls, ids, skip_nonexistents=False):
        data = {'file_ids': ','.join(map(str, ids))}
        if skip_nonexistents:
            data['skip_nonexistents'] = 1
        return cls.client.request('/files/delete', method='POST', data=data)

    def move(self, parent_id):
        return self.client.request('/files/move', method='POST',
                                   data={'file_ids': str(self.id), 'parent_id': str(parent_id)})

    def rename(self, name):
        return self.client.request('/files/rename', method='POST',
                                   data={'file_id': str(self.id), 'name': str(name)})

    @classmethod
    def create_folder(cls, name, parent_id=0):
        r = cls.client.request('/files/create-folder', method='POST',
                               data={'name': name, 'parent_id': str(parent_id)})
        f = r['file']
        return cls(f)


class _Transfer(_BaseResource):

    @classmethod
    def list(cls):
        """ List all transfers """

        d = cls.client.request('/transfers/list')
        transfers = d['transfers']
        return [cls(t) for t in transfers]

    @classmethod
    def get(cls, id):
        """ Get transfer details """
        d = cls.client.request('/transfers/%i' % id, method='GET')
        t = d['transfer']
        return cls(t)

    @classmethod
    def add_url(cls, url, parent_id=0, callback_url=None):
        """ Add new transfer from URI"""

        data = {'url': url, 'save_parent_id': parent_id}
        if callback_url:
            data['callback_url'] = callback_url

        d = cls.client.request('/transfers/add', method='POST', data=data)
        return cls(d['transfer'])

    @classmethod
    def add_torrent(cls, path, parent_id=0, callback_url=None):
        params = {'torrent': 'true'}
        data = {'parent_id': parent_id}
        if callback_url:
            data['callback_url'] = callback_url

        with io.open(path, 'rb') as f:
            files = {'file': f}
            d = cls.client.request(UPLOAD_URL, method='POST', params=params, data=data, files=files)

        return cls(d['transfer'])

    @classmethod
    def clean(cls):
        """ Clean finished transfers"""

        return cls.client.request('/transfers/clean', method='POST')

    def cancel(self):
        """ Cancel or remove  transfers"""

        return self.client.request('/transfers/cancel',
                                   method='POST',
                                   data={'transfer_ids': self.id})

    @classmethod
    def cancel_multi(cls, ids):
        return cls.client.request('/transfers/cancel',
                                  method='POST',
                                  data={'transfer_ids': ','.join(map(str, ids))})


class _Account(_BaseResource):

    @classmethod
    def info(cls):
        """ Get Account info"""
        return cls.client.request('/account/info', method='GET')

    @classmethod
    def settings(cls):
        """ Get account settings"""
        return cls.client.request('/account/settings', method='GET')


class _Subtitle(_BaseResource):

    def __init__(self, resource_dict, file_id):
        self.__dict__.update(resource_dict)
        self.file_id = file_id

    def get_download_link(self):
        path = '/files/%d/subtitles/%s' % (self.file_id, self.key)
        params = {}

        response = self.client.request(path, method='HEAD', params=params, raw=True, allow_redirects=False)

        if str(response.status_code)[0] == '2':
            return response.url
        elif response.status_code == 302:
            return response.headers['Location']

    def download(self, dest):
        name = _str(self.name)
        path = self.get_download_link()

        filepath = os.path.join(dest, name)
        logger.info('downloading subtitle file to: %s', filepath)
        response = self.client.request(path, method='GET', raw=True)
        with io.open(filepath, 'wb') as f:
            f.write(response.content)

        return filepath


# Due to a nasty bug in datetime module, datetime.strptime calls
# are not thread-safe and can throw a TypeError. Details: https://bugs.python.org/issue7980
# Here we are implementing simple RFC3339 parser which is used in Put.io APIv2.
def strptime(date):
    """Returns datetime object from the given date, which is in a specific format: YYYY-MM-ddTHH:mm:ss"""
    d = {
        'year': date[0:4],
        'month': date[5:7],
        'day': date[8:10],
        'hour': date[11:13],
        'minute': date[14:16],
        'second': date[17:],
    }

    d = dict((k, int(v)) for k, v in d.items())

    return datetime(**d)


def _str(s):
    """Python 3 compatibility function for converting to str."""
    try:
        if isinstance(s, unicode):
            return s.encode('utf-8', 'replace')
    except NameError:
        pass
    return s



import re
from urllib.parse import urljoin

from sickchill import logger
from sickchill.helper.common import convert_size
from sickchill.oldbeard import helpers, tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("Newpct")

        self.onlyspasearch = None

        self.url = "http://www.newpct.com"
        self.urls = {"search": urljoin(self.url, "index.php")}

        self.cache = tvcache.TVCache(self, min_time=20)

    def search(self, search_strings, age=0, ep_obj=None):
        """
        Search query:
        http://www.newpct.com/index.php?l=doSearch&q=fringe&category_=All&idioma_=1&bus_de_=All
        q => Show name
        category_ = Category 'Shows' (767)
        idioma_ = Language Spanish (1), All
        bus_de_ = Date from (All, mes, semana, ayer, hoy)
        """
        results = []

        # Only search if user conditions are true
        lang_info = "" if not ep_obj or not ep_obj.show else ep_obj.show.lang

        search_params = {"l": "doSearch", "q": "", "category_": "All", "idioma_": 1, "bus_de_": "All"}

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            if self.onlyspasearch:
                search_params["idioma_"] = 1
            else:
                search_params["idioma_"] = "All"

            # Only search if user conditions are true
            if self.onlyspasearch and lang_info != "es" and mode != "RSS":
                logger.debug("Show info is not spanish, skipping provider search")
                continue

            search_params["bus_de_"] = "All" if mode != "RSS" else "semana"

            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_params["q"] = search_string

                data = self.get_url(self.urls["search"], params=search_params, returns="text")
                if not data:
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find("table", id="categoryTable")
                    torrent_rows = torrent_table("tr") if torrent_table else []

                    # Continue only if at least one Release is found
                    if len(torrent_rows) < 3:  # Headers + 1 Torrent + Pagination
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    # 'Fecha', 'Título', 'Tamaño', ''
                    # Date, Title, Size
                    labels = [label.get_text(strip=True) for label in torrent_rows[0]("th")]
                    for row in torrent_rows[1:-1]:
                        try:
                            cells = row("td")

                            torrent_row = row.find("a")
                            download_url = torrent_row.get("href", "")
                            title = self._processTitle(torrent_row.get("title", ""), download_url)
                            if not all([title, download_url]):
                                continue

                            # Provider does not provide seeders/leechers
                            seeders = 1
                            leechers = 0
                            # 2 is the 'Tamaño' column.
                            torrent_size = cells[2].get_text(strip=True)

                            size = convert_size(torrent_size) or -1
                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug(_("Found result: ") + f"{title}")

                            items.append(item)
                        except (AttributeError, TypeError):
                            continue

            results += items

        return results

    def get_url(self, url, post_data=None, params=None, timeout=30, **kwargs):
        """
        returns='content' when trying access to torrent info (For calling torrent client). Previously we must parse
        the URL to get torrent file
        """
        trickery = kwargs.pop("returns", "")
        if trickery == "content":
            kwargs["returns"] = "text"
            data = super().get_url(url, post_data=post_data, params=params, timeout=timeout, **kwargs)
            url = re.search(r"http://tumejorserie.com/descargar/.+\.torrent", data, re.DOTALL).group()
            url = urljoin(self.url, url.rsplit("=", 1)[-1])

        kwargs["returns"] = trickery
        return super().get_url(url, post_data=post_data, params=params, timeout=timeout, **kwargs)

    def download_result(self, result):
        """
        Save the result to disk.
        """

        # check for auth
        if not self.login():
            return False

        urls, filename = self._make_url(result)

        for url in urls:
            # Search results don't return torrent files directly, it returns show sheets so we must parse showSheet to access torrent.
            data = self.get_url(url, returns="text")
            url_torrent = re.search(r"http://tumejorserie.com/descargar/.+\.torrent", data, re.DOTALL).group()

            if url_torrent.startswith("http"):
                self.headers.update({"Referer": "/".join(url_torrent.split("/")[:3]) + "/"})

            logger.info("Downloading a result from {0}".format(url))

            if helpers.download_file(url_torrent, filename, session=self.session, headers=self.headers):
                if self._verify_download(filename):
                    logger.info("Saved result to {0}".format(filename))
                    return True
                else:
                    logger.warning("Could not download {0}".format(url))
                    helpers.remove_file_failed(filename)

        if urls:
            logger.warning("Failed to download any results")

        return False

    @staticmethod
    def _processTitle(title, url):
        # Remove 'Mas informacion sobre ' literal from title
        title = title[22:]
        title = re.sub(r"[ ]{2,}", " ", title, flags=re.I)

        # Quality - Use re module to avoid case sensitive problems with replace
        title = re.sub(r"\[HDTV 1080p?[^\[]*]", "1080p HDTV x264", title, flags=re.I)
        title = re.sub(r"\[HDTV 720p?[^\[]*]", "720p HDTV x264", title, flags=re.I)
        title = re.sub(r"\[ALTA DEFINICION 720p?[^\[]*]", "720p HDTV x264", title, flags=re.I)
        title = re.sub(r"\[HDTV]", "HDTV x264", title, flags=re.I)
        title = re.sub(r"\[DVD[^\[]*]", "DVDrip x264", title, flags=re.I)
        title = re.sub(r"\[BluRay 1080p?[^\[]*]", "1080p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BluRay Rip 1080p?[^\[]*]", "1080p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BluRay Rip 720p?[^\[]*]", "720p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BluRay MicroHD[^\[]*]", "1080p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[MicroHD 1080p?[^\[]*]", "1080p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BLuRay[^\[]*]", "720p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BRrip[^\[]*]", "720p BluRay x264", title, flags=re.I)
        title = re.sub(r"\[BDrip[^\[]*]", "720p BluRay x264", title, flags=re.I)

        # detect hdtv/bluray by url
        # hdtv 1080p example url: http://www.newpct.com/descargar-seriehd/foo/capitulo-610/hdtv-1080p-ac3-5-1/
        # hdtv 720p example url: http://www.newpct.com/descargar-seriehd/foo/capitulo-26/hdtv-720p-ac3-5-1/
        # hdtv example url: http://www.newpct.com/descargar-serie/foo/capitulo-214/hdtv/
        # bluray compilation example url: http://www.newpct.com/descargar-seriehd/foo/capitulo-11/bluray-1080p/
        title_hdtv = re.search(r"HDTV", title, flags=re.I)
        title_720p = re.search(r"720p", title, flags=re.I)
        title_1080p = re.search(r"1080p", title, flags=re.I)
        title_x264 = re.search(r"x264", title, flags=re.I)
        title_bluray = re.search(r"bluray", title, flags=re.I)
        title_serie_hd = re.search(r"descargar-seriehd", title, flags=re.I)
        url_hdtv = re.search(r"HDTV", url, flags=re.I)
        url_720p = re.search(r"720p", url, flags=re.I)
        url_1080p = re.search(r"1080p", url, flags=re.I)
        url_bluray = re.search(r"bluray", url, flags=re.I)

        if not title_hdtv and url_hdtv:
            title += " HDTV"
            if not title_x264:
                title += " x264"
        if not title_bluray and url_bluray:
            title += " BluRay"
            if not title_x264:
                title += " x264"
        if not title_1080p and url_1080p:
            title += " 1080p"
            title_1080p = True
        if not title_720p and url_720p:
            title += " 720p"
            title_720p = True
        if not (title_720p or title_1080p) and title_serie_hd:
            title += " 720p"

        # Language
        title = re.sub(r"\[Spanish[^\[]*]", "SPANISH AUDIO", title, flags=re.I)
        title = re.sub(r"\[Castellano[^\[]*]", "SPANISH AUDIO", title, flags=re.I)
        title = re.sub(r"\[Español[^\[]*]", "SPANISH AUDIO", title, flags=re.I)
        title = re.sub(r"\[AC3 5\.1 Español[^\[]*]", "SPANISH AUDIO", title, flags=re.I)

        if re.search(r"\[V.O.[^\[]*]", title, flags=re.I):
            title += "-NEWPCTVO"
        else:
            title += "-NEWPCT"

        return title.strip()



import glob

override = True
test_text = """
    def test_timeago_%s(self):
        locale = '%s'
    
        date = datetime.datetime.now()
        now = date + datetime.timedelta(seconds=2)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=10)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=60)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=60 * 3.4)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=3600)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=3600 * 2)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=86400)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=86400 * 4.5)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=86400 * 7)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=2592000)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=86400 * 31)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=2592000 * 3.5)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=31536000)
        self.assertEqual(timeago.format(date, now, locale), "%s")
    
        now = date + datetime.timedelta(seconds=31536000 * 2.1)
        self.assertEqual(timeago.format(date, now, locale), "%s")"""

deltas = [10, 3, 2, 4, 4, 3, 2]

test_functions = []

for filename in glob.glob('./*py'):
    with open(filename) as f:
        content = f.read()
        # We need to use a bit of reflexion to load the module dynamically
        if not 'generated by @lolobosse script' in content:
            continue
        locale_name = filename[2:-3]
        print('Generating tests for %s' % locale_name)
        if "import_locale" or "create_tests" in filename:
            continue
        try:
            values = __import__(locale_name).LOCALE
        except:
            print('No locale found for %s' % locale_name)
            continue
        args = [locale_name, locale_name]
        for index, line in enumerate(values):
            if index % 2 == 1:
                # With format
                args.append(line[0] % str(deltas[index/2]))
            else:
                args.append(line[0])
        result = test_text % tuple(args)
        test_functions.append(result)

with open('../../../test/testcase.py', 'r') as input:
    content = input.readlines()
    line_number = 0
    for index, line in enumerate(content):
        if '__name__' in line:
            line_number = index
    content.insert(line_number - 2, ("\n"*2).join(test_functions))
    text = "".join(content)
    input.close()

if not override:
    with open('../../../test/testcase2.py', 'w+') as result:
        result.write(text)
        result.close()
else:
    open('../../../test/testcase.py', 'w').close()
    with open('../../../test/testcase.py', 'w') as original:
        original.write(text)
        original.close()


import re
import traceback
from urllib.parse import urljoin

import validators
from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("IPTorrents")
        self.enable_cookies = True

        self.username = None
        self.password = None
        self.freeleech = False
        self.minseed = 0
        self.minleech = 0
        self.custom_url = None

        self.cache = tvcache.TVCache(self, min_time=10)  # Only poll IPTorrents every 10 minutes max

        self.urls = {
            "base_url": "https://iptorrents.eu",
            "login": "https://iptorrents.eu/take_login.php",
            "search": "https://iptorrents.eu/t?%s%s&q=%s&qf=#torrents",
        }

        self.url = self.urls["base_url"]

        self.categories = "73=&60="

    def login(self):
        cookie_dict = dict_from_cookiejar(self.session.cookies)
        if cookie_dict.get("uid") and cookie_dict.get("pass"):
            return True

        if self.cookies:
            success, status = self.add_cookies_from_ui()
            if not success:
                logger.info(status)
                return False

        login_params = {"username": self.username, "password": self.password, "login": "submit"}

        if self.custom_url:
            if not validators.url(self.custom_url):
                logger.warning("Invalid custom url: {0}".format(self.custom_url))
                return False

        # Get the index, redirects to login
        data = self.get_url(self.custom_url or self.url, returns="text")
        if not data:
            logger.warning("Unable to connect to provider")
            return False

        with BS4Parser(data, "html5lib") as html:
            action = html.find("form", {"action": re.compile(r".*login.*")}).get("action")
            if not action:
                logger.warning("Could not find the login form. Try adding cookies instead")
                return False

        response = self.get_url(urljoin(self.custom_url or self.url, action), post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        # Invalid username and password combination
        if re.search("Invalid username and password combination", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        # You tried too often, please try again after 2 hours!
        if re.search("You tried too often", response):
            logger.warning("You tried too often, please try again after 2 hours! Disable IPTorrents for at least 2 hours")
            return False

        # Captcha!
        if re.search("Captcha verification failed.", response):
            logger.warning("Stupid captcha")
            return False

        return True

    def search(self, search_params, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        freeleech = "&free=on" if self.freeleech else ""

        for mode in search_params:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in search_params[mode]:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                # URL with 50 tv-show results, or max 150 if adjusted in IPTorrents profile
                search_url = self.urls["search"] % (self.categories, freeleech, search_string)
                search_url += ";o=seeders" if mode != "RSS" else ""

                if self.custom_url:
                    if not validators.url(self.custom_url):
                        logger.warning("Invalid custom url: {0}".format(self.custom_url))
                        return results
                    search_url = urljoin(self.custom_url, search_url.split(self.url)[1])

                data = self.get_url(search_url, returns="text")
                if not data:
                    continue

                try:
                    data = re.sub(r"(?im)<button.+?</button>", "", data, 0)
                    with BS4Parser(data, "html5lib") as html:
                        if not html:
                            logger.debug("No data returned from provider")
                            continue

                        if html.find(text="No Torrents Found!"):
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        torrent_table = html.find("table", id="torrents")
                        torrents = torrent_table("tr") if torrent_table else []

                        # Continue only if one Release is found
                        if not torrents or len(torrents) < 2:
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        for result in torrents[1:]:
                            try:
                                title = result("td")[1].find("a").text
                                download_url = urljoin(search_url, result("td")[3].find("a")["href"])
                                seeders = int(result("td")[7].text)
                                leechers = int(result("td")[8].text)
                                torrent_size = result("td")[5].text
                                size = convert_size(torrent_size) or -1
                            except (AttributeError, TypeError, KeyError):
                                continue

                            if not all([title, download_url]):
                                continue

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                except Exception as e:
                    logger.exception("Failed parsing provider. Error: {0!r}".format(str(e)))
                    logger.exception(traceback.format_exc())

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



# Author: Marvin Pinto <me@marvinp.ca>
# Author: Dennis Lutter <lad1337@gmail.com>

from sickchill import logger, settings
from sickchill.oldbeard.common import (
    NOTIFY_DOWNLOAD,
    NOTIFY_GIT_UPDATE,
    NOTIFY_GIT_UPDATE_TEXT,
    NOTIFY_LOGIN,
    NOTIFY_LOGIN_TEXT,
    NOTIFY_SNATCH,
    NOTIFY_SUBTITLE_DOWNLOAD,
    notifyStrings,
)
from sickchill.oldbeard.helpers import getURL, make_session


class Notifier(object):
    """
    Use Telegram to send notifications

    https://telegram.org/
    """

    def __init__(self):
        self.session = make_session()

    def test_notify(self, id=None, api_key=None):
        """
        Send a test notification
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token
        :returns: the notification
        """
        return self._notify_telegram("Test", "This is a test notification from SickChill", id, api_key, force=True)

    def _send_telegram_msg(self, title, msg, id=None, api_key=None):
        """
        Sends a Telegram notification

        :param title: The title of the notification to send
        :param msg: The message string to send
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token

        :returns: True if the message succeeded, False otherwise
        """
        logger.debug("Telegram in use with API KEY: {0}".format(api_key))
        params = {"chat_id": id or settings.TELEGRAM_ID, "text": f"{title} : {msg}"}
        response = getURL(f"https://api.telegram.org/bot{api_key or settings.TELEGRAM_APIKEY}/sendMessage", params=params, session=self.session, returns="json")
        message = ("Telegram message sent successfully.", "Sending Telegram message failed, check the log")[response is None]
        logger.info(message)
        return response is not None, message

    def notify_snatch(self, ep_name, title=notifyStrings[NOTIFY_SNATCH]):
        """
        Sends a Telegram notification when an episode is snatched

        :param ep_name: The name of the episode snatched
        :param title: The title of the notification to send
        """
        if settings.TELEGRAM_NOTIFY_ONSNATCH:
            self._notify_telegram(title, ep_name)

    def notify_download(self, ep_name, title=notifyStrings[NOTIFY_DOWNLOAD]):
        """
        Sends a Telegram notification when an episode is downloaded

        :param ep_name: The name of the episode downloaded
        :param title: The title of the notification to send
        """
        if settings.TELEGRAM_NOTIFY_ONDOWNLOAD:
            self._notify_telegram(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=notifyStrings[NOTIFY_SUBTITLE_DOWNLOAD]):
        """
        Sends a Telegram notification when subtitles for an episode are downloaded

        :param ep_name: The name of the episode subtitles were downloaded for
        :param lang: The language of the downloaded subtitles
        :param title: The title of the notification to send
        """
        if settings.TELEGRAM_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notify_telegram(title, "{0}: {1}".format(ep_name, lang))

    def notify_git_update(self, new_version="??"):
        """
        Sends a Telegram notification for git updates

        :param new_version: The new version available from git
        """
        if settings.USE_TELEGRAM:
            update_text = notifyStrings[NOTIFY_GIT_UPDATE_TEXT]
            title = notifyStrings[NOTIFY_GIT_UPDATE]
            self._notify_telegram(title, update_text + new_version)

    def notify_login(self, ipaddress=""):
        """
        Sends a Telegram notification on login

        :param ipaddress: The ip address the login is originating from
        """
        if settings.USE_TELEGRAM:
            update_text = notifyStrings[NOTIFY_LOGIN_TEXT]
            title = notifyStrings[NOTIFY_LOGIN]
            self._notify_telegram(title, update_text.format(ipaddress))

    def _notify_telegram(self, title, message, id=None, api_key=None, force=False):
        """
        Sends a Telegram notification

        :param title: The title of the notification to send
        :param message: The message string to send
        :param id: The Telegram user/group id to send the message to
        :param api_key: Your Telegram bot API token
        :param force: Enforce sending, for instance for testing

        :returns: the message to send
        """

        if not (force or settings.USE_TELEGRAM):
            logger.debug("Notification for Telegram not enabled, skipping this notification")
            return False, "Disabled"

        logger.debug("Sending a Telegram message for {0}".format(message))

        return self._send_telegram_msg(title, message, id, api_key)
